import pytest

from django.core.exceptions import ImproperlyConfigured
from django.db import connection, models

from psqlextra.backend.schema import PostgresSchemaEditor
from psqlextra.types import PostgresPartitioningMethod

from . import db_introspection
from .fake_model import define_fake_partitioned_model


def test_schema_editor_create_delete_partitioned_model_range():
    """Tests whether creating a partitioned model and adding a list partition
    to it using the :see:PostgresSchemaEditor works."""

    method = PostgresPartitioningMethod.RANGE
    key = ["timestamp"]

    model = define_fake_partitioned_model(
        {"name": models.TextField(), "timestamp": models.DateTimeField()},
        {"method": method, "key": key},
    )

    schema_editor = PostgresSchemaEditor(connection)
    schema_editor.create_partitioned_model(model)

    schema_editor.add_range_partition(model, "pt1", "2019-01-01", "2019-02-01")

    table = db_introspection.get_partitioned_table(model._meta.db_table)
    assert table.name == model._meta.db_table
    assert table.method == method
    assert table.key == key
    assert table.partitions[0].full_name == model._meta.db_table + "_pt1"

    schema_editor.delete_partitioned_model(model)

    table = db_introspection.get_partitioned_table(model._meta.db_table)
    assert not table

    partitions = db_introspection.get_partitions(model._meta.db_table)
    assert len(partitions) == 0


def test_schema_editor_create_delete_partitioned_model_list():
    """Tests whether creating a partitioned model and adding a range partition
    to it using the :see:PostgresSchemaEditor works."""

    method = PostgresPartitioningMethod.LIST
    key = ["category"]

    model = define_fake_partitioned_model(
        {"name": models.TextField(), "category": models.TextField()},
        {"method": method, "key": key},
    )

    schema_editor = PostgresSchemaEditor(connection)
    schema_editor.create_partitioned_model(model)

    schema_editor.add_list_partition(model, "pt1", ["car", "boat"])

    table = db_introspection.get_partitioned_table(model._meta.db_table)
    assert table.name == model._meta.db_table
    assert table.method == method
    assert table.key == key
    assert table.partitions[0].full_name == model._meta.db_table + "_pt1"

    schema_editor.delete_partitioned_model(model)

    table = db_introspection.get_partitioned_table(model._meta.db_table)
    assert not table

    partitions = db_introspection.get_partitions(model._meta.db_table)
    assert len(partitions) == 0


def test_schema_editor_create_delete_partitioned_model_default():
    """Tests whether creating a partitioned model and adding a default
    partition to it using the :see:PostgresSchemaEditor works."""

    method = PostgresPartitioningMethod.LIST
    key = ["category"]

    model = define_fake_partitioned_model(
        {"name": models.TextField(), "category": models.TextField()},
        {"method": method, "key": key},
    )

    schema_editor = PostgresSchemaEditor(connection)
    schema_editor.create_partitioned_model(model)

    schema_editor.add_default_partition(model, "default")

    table = db_introspection.get_partitioned_table(model._meta.db_table)
    assert table.name == model._meta.db_table
    assert table.method == method
    assert table.key == key
    assert table.partitions[0].full_name == model._meta.db_table + "_default"

    schema_editor.delete_partitioned_model(model)

    table = db_introspection.get_partitioned_table(model._meta.db_table)
    assert not table

    partitions = db_introspection.get_partitions(model._meta.db_table)
    assert len(partitions) == 0


def test_schema_editor_create_partitioned_model_no_method():
    """Tests whether its possible to create a partitioned model without
    explicitly setting a partitioning method.

    The default is "range" so setting one explicitely should not be
    needed.
    """

    model = define_fake_partitioned_model(
        {"name": models.TextField(), "timestamp": models.DateTimeField()},
        {"key": ["timestamp"]},
    )

    schema_editor = PostgresSchemaEditor(connection)
    schema_editor.create_partitioned_model(model)

    pt = db_introspection.get_partitioned_table(model._meta.db_table)
    assert pt.method == PostgresPartitioningMethod.RANGE
    assert len(pt.partitions) == 0


def test_schema_editor_create_partitioned_model_no_key():
    """Tests whether trying to create a partitioned model without a
    partitioning key raises :see:ImproperlyConfigured as its not possible to
    create a partitioned model without one and we cannot have a sane
    default."""

    model = define_fake_partitioned_model(
        {"name": models.TextField(), "timestamp": models.DateTimeField()},
        {"method": PostgresPartitioningMethod.RANGE},
    )

    schema_editor = PostgresSchemaEditor(connection)

    with pytest.raises(ImproperlyConfigured):
        schema_editor.create_partitioned_model(model)


def test_schema_editor_add_range_partition():
    """Tests whether adding a range partition works."""

    model = define_fake_partitioned_model(
        {"name": models.TextField(), "timestamp": models.DateTimeField()},
        {"key": ["timestamp"]},
    )

    schema_editor = PostgresSchemaEditor(connection)
    schema_editor.create_partitioned_model(model)

    schema_editor.add_range_partition(
        model,
        name="mypartition",
        from_values="2019-1-1",
        to_values="2019-2-1",
        comment="test",
    )

    table = db_introspection.get_partitioned_table(model._meta.db_table)
    assert len(table.partitions) == 1
    assert table.partitions[0].name == "mypartition"
    assert (
        table.partitions[0].full_name == f"{model._meta.db_table}_mypartition"
    )
    assert table.partitions[0].comment == "test"

    schema_editor.delete_partition(model, "mypartition")
    table = db_introspection.get_partitioned_table(model._meta.db_table)
    assert len(table.partitions) == 0


def test_schema_editor_add_list_partition():
    """Tests whether adding a list partition works."""

    model = define_fake_partitioned_model(
        {"name": models.TextField()},
        {"method": PostgresPartitioningMethod.LIST, "key": ["name"]},
    )

    schema_editor = PostgresSchemaEditor(connection)
    schema_editor.create_partitioned_model(model)

    schema_editor.add_list_partition(
        model, name="mypartition", values=["1"], comment="test"
    )

    table = db_introspection.get_partitioned_table(model._meta.db_table)
    assert len(table.partitions) == 1
    assert table.partitions[0].name == "mypartition"
    assert (
        table.partitions[0].full_name == f"{model._meta.db_table}_mypartition"
    )
    assert table.partitions[0].comment == "test"

    schema_editor.delete_partition(model, "mypartition")
    table = db_introspection.get_partitioned_table(model._meta.db_table)
    assert len(table.partitions) == 0


@pytest.mark.parametrize(
    "method,key",
    [
        (PostgresPartitioningMethod.RANGE, ["timestamp"]),
        (PostgresPartitioningMethod.LIST, ["name"]),
    ],
)
def test_schema_editor_add_default_partition(method, key):
    model = define_fake_partitioned_model(
        {"name": models.TextField(), "timestamp": models.DateTimeField()},
        {"method": method, "key": key},
    )

    schema_editor = PostgresSchemaEditor(connection)
    schema_editor.create_partitioned_model(model)

    schema_editor.add_default_partition(
        model, name="mypartition", comment="test"
    )

    table = db_introspection.get_partitioned_table(model._meta.db_table)
    assert len(table.partitions) == 1
    assert table.partitions[0].name == "mypartition"
    assert (
        table.partitions[0].full_name == f"{model._meta.db_table}_mypartition"
    )
    assert table.partitions[0].comment == "test"

    schema_editor.delete_partition(model, "mypartition")
    table = db_introspection.get_partitioned_table(model._meta.db_table)
    assert len(table.partitions) == 0



from collections.abc import Iterable

from django.core.exceptions import SuspiciousOperation
from django.db.models import Model
from django.db.models.fields.related import RelatedField
from django.db.models.sql.compiler import SQLInsertCompiler, SQLUpdateCompiler
from django.db.utils import ProgrammingError

from .expressions import HStoreValue
from .types import ConflictAction


class PostgresUpdateCompiler(SQLUpdateCompiler):
    """Compiler for SQL UPDATE statements that allows us to use expressions
    inside HStore values.

    Like:

        .update(name=dict(en=F('test')))
    """

    def as_sql(self):
        self._prepare_query_values()
        return super().as_sql()

    def _prepare_query_values(self):
        """Extra prep on query values by converting dictionaries into.

        :see:HStoreValue expressions.

        This allows putting expressions in a dictionary. The
        :see:HStoreValue will take care of resolving the expressions
        inside the dictionary.
        """

        if not self.query.values:
            return

        new_query_values = []
        for field, model, val in self.query.values:
            if not isinstance(val, dict):
                new_query_values.append((field, model, val))
                continue

            if not self._does_dict_contain_expression(val):
                new_query_values.append((field, model, val))
                continue

            expression = HStoreValue(dict(val))
            new_query_values.append((field, model, expression))

        self.query.values = new_query_values

    @staticmethod
    def _does_dict_contain_expression(data: dict) -> bool:
        """Gets whether the specified dictionary contains any expressions that
        need to be resolved."""

        for value in data.values():
            if hasattr(value, "resolve_expression"):
                return True

            if hasattr(value, "as_sql"):
                return True

        return False


class PostgresInsertCompiler(SQLInsertCompiler):
    """Compiler for SQL INSERT statements."""

    def __init__(self, *args, **kwargs):
        """Initializes a new instance of :see:PostgresInsertCompiler."""

        super().__init__(*args, **kwargs)
        self.qn = self.connection.ops.quote_name

    def as_sql(self, return_id=False):
        """Builds the SQL INSERT statement."""

        queries = [
            self._rewrite_insert(sql, params, return_id)
            for sql, params in super().as_sql()
        ]

        return queries

    def execute_sql(self, return_id=False):
        # execute all the generate queries
        with self.connection.cursor() as cursor:
            rows = []
            for sql, params in self.as_sql(return_id):
                cursor.execute(sql, params)
                try:
                    rows.extend(cursor.fetchall())
                except ProgrammingError:
                    pass

        # create a mapping between column names and column value
        return [
            {
                column.name: row[column_index]
                for column_index, column in enumerate(cursor.description)
                if row
            }
            for row in rows
        ]

    def _rewrite_insert(self, sql, params, return_id=False):
        """Rewrites a formed SQL INSERT query to include the ON CONFLICT
        clause.

        Arguments:
            sql:
                The SQL INSERT query to rewrite.

            params:
                The parameters passed to the query.

            returning:
                What to put in the `RETURNING` clause
                of the resulting query.

        Returns:
            A tuple of the rewritten SQL query and new params.
        """

        returning = (
            self.qn(self.query.model._meta.pk.attname) if return_id else "*"
        )

        return self._rewrite_insert_on_conflict(
            sql, params, self.query.conflict_action.value, returning
        )

    def _rewrite_insert_on_conflict(
        self, sql, params, conflict_action: ConflictAction, returning
    ):
        """Rewrites a normal SQL INSERT query to add the 'ON CONFLICT'
        clause."""

        update_columns = ", ".join(
            [
                "{0} = EXCLUDED.{0}".format(self.qn(field.column))
                for field in self.query.update_fields
            ]
        )

        # build the conflict target, the columns to watch
        # for conflicts
        conflict_target = self._build_conflict_target()
        index_predicate = self.query.index_predicate

        sql_template = (
            "{insert} ON CONFLICT {conflict_target} DO {conflict_action}"
        )

        if index_predicate:
            sql_template = "{insert} ON CONFLICT {conflict_target} WHERE {index_predicate} DO {conflict_action}"

        if conflict_action == "UPDATE":
            sql_template += " SET {update_columns}"

        sql_template += " RETURNING {returning}"

        return (
            sql_template.format(
                insert=sql,
                conflict_target=conflict_target,
                conflict_action=conflict_action,
                update_columns=update_columns,
                returning=returning,
                index_predicate=index_predicate,
            ),
            params,
        )

    def _build_conflict_target(self):
        """Builds the `conflict_target` for the ON CONFLICT clause."""

        if not isinstance(self.query.conflict_target, Iterable):
            raise SuspiciousOperation(
                (
                    "%s is not a valid conflict target, specify "
                    "a list of column names, or tuples with column "
                    "names and hstore key."
                )
                % str(self.query.conflict_target)
            )

        conflict_target = self._build_conflict_target_by_index()
        if conflict_target:
            return conflict_target

        return self._build_conflict_target_by_fields()

    def _build_conflict_target_by_fields(self):
        """Builds the `conflict_target` for the ON CONFLICT clauses by matching
        the fields specified in the specified conflict target against the
        model's fields.

        This requires some special handling because the fields names
        might not be same as the column names.
        """

        conflict_target = []

        for field_name in self.query.conflict_target:
            self._assert_valid_field(field_name)

            # special handling for hstore keys
            if isinstance(field_name, tuple):
                conflict_target.append(
                    "(%s->'%s')"
                    % (self._format_field_name(field_name), field_name[1])
                )
            else:
                conflict_target.append(self._format_field_name(field_name))

        return "(%s)" % ",".join(conflict_target)

    def _build_conflict_target_by_index(self):
        """Builds the `conflict_target` for the ON CONFLICT clause by trying to
        find an index that matches the specified conflict target on the query.

        Conflict targets must match some unique constraint, usually this
        is a `UNIQUE INDEX`.
        """

        matching_index = next(
            (
                index
                for index in self.query.model._meta.indexes
                if list(index.fields) == list(self.query.conflict_target)
            ),
            None,
        )

        if not matching_index:
            return None

        with self.connection.schema_editor() as schema_editor:
            stmt = matching_index.create_sql(self.query.model, schema_editor)
            return "(%s)" % stmt.parts["columns"]

    def _get_model_field(self, name: str):
        """Gets the field on a model with the specified name.

        Arguments:
            name:
                The name of the field to look for.

                This can be both the actual field name, or
                the name of the column, both will work :)

        Returns:
            The field with the specified name or None if
            no such field exists.
        """

        field_name = self._normalize_field_name(name)

        # 'pk' has special meaning and always refers to the primary
        # key of a model, we have to respect this de-facto standard behaviour
        if field_name == "pk" and self.query.model._meta.pk:
            return self.query.model._meta.pk

        for field in self.query.model._meta.local_concrete_fields:
            if field.name == field_name or field.column == field_name:
                return field

        return None

    def _format_field_name(self, field_name) -> str:
        """Formats a field's name for usage in SQL.

        Arguments:
            field_name:
                The field name to format.

        Returns:
            The specified field name formatted for
            usage in SQL.
        """

        field = self._get_model_field(field_name)
        return self.qn(field.column)

    def _format_field_value(self, field_name) -> str:
        """Formats a field's value for usage in SQL.

        Arguments:
            field_name:
                The name of the field to format
                the value of.

        Returns:
            The field's value formatted for usage
            in SQL.
        """

        field_name = self._normalize_field_name(field_name)
        field = self._get_model_field(field_name)

        value = getattr(self.query.objs[0], field.attname)

        if isinstance(field, RelatedField) and isinstance(value, Model):
            value = value.pk

        return SQLInsertCompiler.prepare_value(
            self,
            field,
            # Note: this deliberately doesn't use `pre_save_val` as we don't
            # want things like auto_now on DateTimeField (etc.) to change the
            # value. We rely on pre_save having already been done by the
            # underlying compiler so that things like FileField have already had
            # the opportunity to save out their data.
            value,
        )

    def _assert_valid_field(self, field_name: str):
        """Asserts that a field with the specified name exists on the model and
        raises :see:SuspiciousOperation if it does not."""

        field_name = self._normalize_field_name(field_name)
        if self._get_model_field(field_name):
            return

        raise SuspiciousOperation(
            (
                "%s is not a valid conflict target, specify "
                "a list of column names, or tuples with column "
                "names and hstore key."
            )
            % str(field_name)
        )

    @staticmethod
    def _normalize_field_name(field_name: str) -> str:
        """Normalizes a field name into a string by extracting the field name
        if it was specified as a reference to a HStore key (as a tuple).

        Arguments:
            field_name:
                The field name to normalize.

        Returns:
            The normalized field name.
        """

        if isinstance(field_name, tuple):
            field_name, _ = field_name

        return field_name



import argparse

from unittest.mock import MagicMock, create_autospec, patch

import pytest

from django.db import models
from django.test import override_settings

from psqlextra.backend.introspection import (
    PostgresIntrospectedPartitionTable,
    PostgresIntrospectedPartitonedTable,
)
from psqlextra.management.commands.pgpartition import Command
from psqlextra.partitioning import PostgresPartitioningManager
from psqlextra.partitioning.config import PostgresPartitioningConfig
from psqlextra.partitioning.partition import PostgresPartition
from psqlextra.partitioning.strategy import PostgresPartitioningStrategy

from .fake_model import define_fake_partitioned_model


@pytest.fixture
def fake_strategy():
    strategy = create_autospec(PostgresPartitioningStrategy)

    strategy.createable_partition = create_autospec(PostgresPartition)
    strategy.createable_partition.name = MagicMock(return_value="tobecreated")
    strategy.to_create = MagicMock(return_value=[strategy.createable_partition])

    strategy.deleteable_partition = create_autospec(PostgresPartition)
    strategy.deleteable_partition.name = MagicMock(return_value="tobedeleted")
    strategy.to_delete = MagicMock(return_value=[strategy.deleteable_partition])

    return strategy


@pytest.fixture
def fake_model(fake_strategy):
    model = define_fake_partitioned_model(
        {"timestamp": models.DateTimeField()}, {"key": ["timestamp"]}
    )

    # consistent model name so snapshot tests work
    model.__name__ = "test"

    # we have to trick the system into thinking the model/table
    # actually exists with one partition (so we can simulate deletions)
    deleteable_partition_name = fake_strategy.deleteable_partition.name()
    mocked_partitioned_table = PostgresIntrospectedPartitonedTable(
        name=model._meta.db_table,
        method=model._partitioning_meta.method,
        key=model._partitioning_meta.key,
        partitions=[
            PostgresIntrospectedPartitionTable(
                name=deleteable_partition_name,
                full_name=f"{model._meta.db_table}_{deleteable_partition_name}",
                comment="psqlextra_auto_partitioned",
            )
        ],
    )

    introspection_package = "psqlextra.backend.introspection"
    introspection_class = f"{introspection_package}.PostgresIntrospection"
    get_partitioned_table_path = f"{introspection_class}.get_partitioned_table"

    with patch(get_partitioned_table_path) as mock:
        mock.return_value = mocked_partitioned_table
        yield model


@pytest.fixture
def fake_partitioning_manager(fake_model, fake_strategy):
    manager = PostgresPartitioningManager(
        [PostgresPartitioningConfig(fake_model, fake_strategy)]
    )

    with override_settings(PSQLEXTRA_PARTITIONING_MANAGER=manager):
        yield manager


@pytest.fixture
def run(capsys):
    def _run(*args):
        parser = argparse.ArgumentParser()

        command = Command()
        command.add_arguments(parser)
        command.handle(**vars(parser.parse_args(args)))

        return capsys.readouterr()

    return _run


@pytest.mark.parametrize("args", ["-d", "--dry"])
def test_management_command_partition_dry_run(
    args, snapshot, run, fake_model, fake_partitioning_manager
):
    """Tests whether the --dry option actually makes it a dry run and does not
    create/delete partitions."""

    config = fake_partitioning_manager.find_config_for_model(fake_model)
    snapshot.assert_match(run(args))

    config.strategy.createable_partition.create.assert_not_called()
    config.strategy.createable_partition.delete.assert_not_called()
    config.strategy.deleteable_partition.create.assert_not_called()
    config.strategy.deleteable_partition.delete.assert_not_called()


@pytest.mark.parametrize("args", ["-y", "--yes"])
def test_management_command_partition_auto_confirm(
    args, snapshot, run, fake_model, fake_partitioning_manager
):
    """Tests whether the --yes option makes it not ask for confirmation before
    creating/deleting partitions."""

    config = fake_partitioning_manager.find_config_for_model(fake_model)
    snapshot.assert_match(run(args))

    config.strategy.createable_partition.create.assert_called_once()
    config.strategy.createable_partition.delete.assert_not_called()
    config.strategy.deleteable_partition.create.assert_not_called()
    config.strategy.deleteable_partition.delete.assert_called_once()


@pytest.mark.parametrize("answer", ["y", "Y", "yes", "YES"])
def test_management_command_partition_confirm_yes(
    answer, monkeypatch, snapshot, run, fake_model, fake_partitioning_manager
):
    """Tests whether the --yes option makes it not ask for confirmation before
    creating/deleting partitions."""

    config = fake_partitioning_manager.find_config_for_model(fake_model)

    monkeypatch.setattr("builtins.input", lambda _: answer)
    snapshot.assert_match(run())

    config.strategy.createable_partition.create.assert_called_once()
    config.strategy.createable_partition.delete.assert_not_called()
    config.strategy.deleteable_partition.create.assert_not_called()
    config.strategy.deleteable_partition.delete.assert_called_once()


@pytest.mark.parametrize("answer", ["n", "N", "no", "No", "NO"])
def test_management_command_partition_confirm_no(
    answer, monkeypatch, snapshot, run, fake_model, fake_partitioning_manager
):
    """Tests whether the --yes option makes it not ask for confirmation before
    creating/deleting partitions."""

    config = fake_partitioning_manager.find_config_for_model(fake_model)

    monkeypatch.setattr("builtins.input", lambda _: answer)
    snapshot.assert_match(run())

    config.strategy.createable_partition.create.assert_not_called()
    config.strategy.createable_partition.delete.assert_not_called()
    config.strategy.deleteable_partition.create.assert_not_called()
    config.strategy.deleteable_partition.delete.assert_not_called()



from contextlib import contextmanager
from typing import List
from unittest import mock

from django.apps import apps
from django.db import connection, migrations
from django.db.migrations.autodetector import MigrationAutodetector
from django.db.migrations.executor import MigrationExecutor
from django.db.migrations.loader import MigrationLoader
from django.db.migrations.questioner import NonInteractiveMigrationQuestioner
from django.db.migrations.state import ProjectState

from psqlextra.backend.schema import PostgresSchemaEditor

from .fake_model import define_fake_model


@contextmanager
def filtered_schema_editor(*filters: List[str]):
    """Gets a schema editor, but filters executed SQL statements based on the
    specified text filters.

    Arguments:
        filters:
            List of strings to filter SQL
            statements on.
    """

    with connection.schema_editor() as schema_editor:
        wrapper_for = schema_editor.execute
        with mock.patch.object(
            PostgresSchemaEditor, "execute", wraps=wrapper_for
        ) as execute:
            filter_results = {}
            yield filter_results

    for filter_text in filters:
        filter_results[filter_text] = [
            call for call in execute.mock_calls if filter_text in str(call)
        ]


def apply_migration(operations, state=None, backwards: bool = False):
    """Executes the specified migration operations using the specified schema
    editor.

    Arguments:
        operations:
            The migration operations to execute.

        state:
            The state state to use during the
            migrations.

        backwards:
            Whether to apply the operations
            in reverse (backwards).
    """

    state = state or migrations.state.ProjectState.from_apps(apps)

    class Migration(migrations.Migration):
        pass

    Migration.operations = operations

    migration = Migration("migration", "tests")
    executor = MigrationExecutor(connection)

    if not backwards:
        executor.apply_migration(state, migration)
    else:
        executor.unapply_migration(state, migration)

    return migration


def make_migration(app_label="tests", from_state=None, to_state=None):
    """Generates migrations based on the specified app's state."""

    app_labels = [app_label]

    loader = MigrationLoader(None, ignore_no_migrations=True)
    loader.check_consistent_history(connection)

    questioner = NonInteractiveMigrationQuestioner(
        specified_apps=app_labels, dry_run=False
    )

    autodetector = MigrationAutodetector(
        from_state or loader.project_state(),
        to_state or ProjectState.from_apps(apps),
        questioner,
    )

    changes = autodetector.changes(
        graph=loader.graph,
        trim_to_apps=app_labels or None,
        convert_apps=app_labels or None,
        migration_name="test",
    )

    changes_for_app = changes.get(app_label)
    if not changes_for_app or len(changes_for_app) == 0:
        return None

    return changes_for_app[0]


@contextmanager
def create_drop_model(field, filters: List[str]):
    """Creates and drops a model with the specified field.

    Arguments:
        field:
            The field to include on the
            model to create and drop.

        filters:
            List of strings to filter
            SQL statements on.
    """

    model = define_fake_model({"title": field})

    with filtered_schema_editor(*filters) as calls:
        apply_migration(
            [
                migrations.CreateModel(
                    model.__name__, fields=[("title", field.clone())]
                ),
                migrations.DeleteModel(model.__name__),
            ]
        )

    yield calls


@contextmanager
def alter_db_table(field, filters: List[str]):
    """Creates a model with the specified field and then renames the database
    table.

    Arguments:
        field:
            The field to include into the
            model.

        filters:
            List of strings to filter
            SQL statements on.
    """

    model = define_fake_model()
    state = migrations.state.ProjectState.from_apps(apps)

    apply_migration(
        [
            migrations.CreateModel(
                model.__name__, fields=[("title", field.clone())]
            )
        ],
        state,
    )

    with filtered_schema_editor(*filters) as calls:
        apply_migration(
            [migrations.AlterModelTable(model.__name__, "NewTableName")], state
        )

    yield calls


@contextmanager
def add_field(field, filters: List[str]):
    """Adds the specified field to a model.

    Arguments:
        field:
            The field to add to a model.

        filters:
            List of strings to filter
            SQL statements on.
    """

    model = define_fake_model()
    state = migrations.state.ProjectState.from_apps(apps)

    apply_migration([migrations.CreateModel(model.__name__, fields=[])], state)

    with filtered_schema_editor(*filters) as calls:
        apply_migration(
            [migrations.AddField(model.__name__, "title", field)], state
        )

    yield calls


@contextmanager
def remove_field(field, filters: List[str]):
    """Removes the specified field from a model.

    Arguments:
        field:
            The field to remove from a model.

        filters:
            List of strings to filter
            SQL statements on.
    """

    model = define_fake_model({"title": field})
    state = migrations.state.ProjectState.from_apps(apps)

    apply_migration(
        [
            migrations.CreateModel(
                model.__name__, fields=[("title", field.clone())]
            )
        ],
        state,
    )

    with filtered_schema_editor(*filters) as calls:
        apply_migration(
            [migrations.RemoveField(model.__name__, "title")], state
        )

    yield calls


@contextmanager
def alter_field(old_field, new_field, filters: List[str]):
    """Alters a field from one state to the other.

    Arguments:
        old_field:
            The field before altering it.

        new_field:
            The field after altering it.

        filters:
            List of strings to filter
            SQL statements on.
    """

    model = define_fake_model({"title": old_field})
    state = migrations.state.ProjectState.from_apps(apps)

    apply_migration(
        [
            migrations.CreateModel(
                model.__name__, fields=[("title", old_field.clone())]
            )
        ],
        state,
    )

    with filtered_schema_editor(*filters) as calls:
        apply_migration(
            [migrations.AlterField(model.__name__, "title", new_field)], state
        )

    yield calls


@contextmanager
def rename_field(field, filters: List[str]):
    """Renames a field from one name to the other.

    Arguments:
        field:
            Field to be renamed.

        filters:
            List of strings to filter
            SQL statements on.
    """

    model = define_fake_model({"title": field})
    state = migrations.state.ProjectState.from_apps(apps)

    apply_migration(
        [
            migrations.CreateModel(
                model.__name__, fields=[("title", field.clone())]
            )
        ],
        state,
    )

    with filtered_schema_editor(*filters) as calls:
        apply_migration(
            [migrations.RenameField(model.__name__, "title", "newtitle")], state
        )

    yield calls



from contextlib import contextmanager
from unittest import mock

from django.db.migrations import (
    AddField,
    AlterField,
    CreateModel,
    DeleteModel,
    RemoveField,
    RenameField,
    RunSQL,
)
from django.db.migrations.autodetector import MigrationAutodetector
from django.db.migrations.operations.base import Operation
from django.db.models import Model

from psqlextra.models import (
    PostgresMaterializedViewModel,
    PostgresPartitionedModel,
    PostgresViewModel,
)

from . import operations

# original `MigrationAutodetector.add_operation`
# function, saved here so the patched version can
# call the original
add_operation = MigrationAutodetector.add_operation


class AddOperationHandler:
    """Handler for when operations are being added to a new migration.

    This is where we intercept operations such as
    :see:CreateModel to replace it with our own.
    """

    def __init__(self, autodetector, app_label, args, kwargs):
        self.autodetector = autodetector
        self.app_label = app_label
        self.args = args
        self.kwargs = kwargs

    def add(self, operation):
        """Adds the specified operation to the list of operations to execute in
        the migration."""

        return add_operation(
            self.autodetector,
            self.app_label,
            operation,
            *self.args,
            **self.kwargs,
        )

    def add_field(self, operation: AddField):
        """Adds the specified :see:AddField operation to the list of operations
        to execute in the migration."""

        return self._transform_view_field_operations(operation)

    def remove_field(self, operation: RemoveField):
        """Adds the specified :see:RemoveField operation to the list of
        operations to execute in the migration."""

        return self._transform_view_field_operations(operation)

    def alter_field(self, operation: AlterField):
        """Adds the specified :see:AlterField operation to the list of
        operations to execute in the migration."""

        return self._transform_view_field_operations(operation)

    def rename_field(self, operation: RenameField):
        """Adds the specified :see:RenameField operation to the list of
        operations to execute in the migration."""

        return self._transform_view_field_operations(operation)

    def _transform_view_field_operations(self, operation: Operation):
        """Transforms operations on fields on a (materialized) view into state
        only operations.

        One cannot add/remove/delete fields on a (materialized) view,
        however, we do want Django's migration system to keep track of
        these kind of changes to the model. The :see:ApplyState
        operation just tells Django the operation was applied without
        actually applying it.
        """

        model = self.autodetector.new_apps.get_model(
            self.app_label, operation.model_name
        )

        if issubclass(model, PostgresViewModel):
            return self.add(operations.ApplyState(state_operation=operation))

        return self.add(operation)

    def add_create_model(self, operation: CreateModel):
        """Adds the specified :see:CreateModel operation to the list of
        operations to execute in the migration."""

        model = self.autodetector.new_apps.get_model(
            self.app_label, operation.name
        )

        if issubclass(model, PostgresPartitionedModel):
            return self.add_create_partitioned_model(model, operation)
        elif issubclass(model, PostgresMaterializedViewModel):
            return self.add_create_materialized_view_model(model, operation)
        elif issubclass(model, PostgresViewModel):
            return self.add_create_view_model(model, operation)

        return self.add(operation)

    def add_delete_model(self, operation: DeleteModel):
        """Adds the specified :see:Deletemodel operation to the list of
        operations to execute in the migration."""

        model = self.autodetector.old_apps.get_model(
            self.app_label, operation.name
        )

        if issubclass(model, PostgresPartitionedModel):
            return self.add_delete_partitioned_model(model, operation)
        elif issubclass(model, PostgresMaterializedViewModel):
            return self.add_delete_materialized_view_model(model, operation)
        elif issubclass(model, PostgresViewModel):
            return self.add_delete_view_model(model, operation)

        return self.add(operation)

    def add_create_partitioned_model(
        self, model: Model, operation: CreateModel
    ):
        """Adds a :see:PostgresCreatePartitionedModel operation to the list of
        operations to execute in the migration."""

        partitioning_options = model._partitioning_meta.original_attrs
        _, args, kwargs = operation.deconstruct()

        self.add(
            operations.PostgresAddDefaultPartition(
                model_name=model.__name__, name="default"
            )
        )

        self.add(
            operations.PostgresCreatePartitionedModel(
                *args, **kwargs, partitioning_options=partitioning_options
            )
        )

    def add_delete_partitioned_model(
        self, model: Model, operation: DeleteModel
    ):
        """Adds a :see:PostgresDeletePartitionedModel operation to the list of
        operations to execute in the migration."""

        _, args, kwargs = operation.deconstruct()
        return self.add(
            operations.PostgresDeletePartitionedModel(*args, **kwargs)
        )

    def add_create_view_model(self, model: Model, operation: CreateModel):
        """Adds a :see:PostgresCreateViewModel operation to the list of
        operations to execute in the migration."""

        view_options = model._view_meta.original_attrs
        _, args, kwargs = operation.deconstruct()

        self.add(
            operations.PostgresCreateViewModel(
                *args, **kwargs, view_options=view_options
            )
        )

    def add_delete_view_model(self, model: Model, operation: DeleteModel):
        """Adds a :see:PostgresDeleteViewModel operation to the list of
        operations to execute in the migration."""

        _, args, kwargs = operation.deconstruct()
        return self.add(operations.PostgresDeleteViewModel(*args, **kwargs))

    def add_create_materialized_view_model(
        self, model: Model, operation: CreateModel
    ):
        """Adds a :see:PostgresCreateMaterializedViewModel operation to the
        list of operations to execute in the migration."""

        view_options = model._view_meta.original_attrs
        _, args, kwargs = operation.deconstruct()

        self.add(
            operations.PostgresCreateMaterializedViewModel(
                *args, **kwargs, view_options=view_options
            )
        )

    def add_delete_materialized_view_model(
        self, model: Model, operation: DeleteModel
    ):
        """Adds a :see:PostgresDeleteMaterializedViewModel operation to the
        list of operations to execute in the migration."""

        _, args, kwargs = operation.deconstruct()
        return self.add(
            operations.PostgresDeleteMaterializedViewModel(*args, **kwargs)
        )


@contextmanager
def patched_autodetector():
    """Patches the standard Django :seee:MigrationAutodetector for the duration
    of the context.

    The patch intercepts the `add_operation` function to
    customize how new operations are added.

    We have to do this because there is no way in Django
    to extend the auto detector otherwise.
    """

    autodetector_module_path = "django.db.migrations.autodetector"
    autodetector_class_path = (
        f"{autodetector_module_path}.MigrationAutodetector"
    )
    add_operation_path = f"{autodetector_class_path}.add_operation"

    def _patched(autodetector, app_label, operation, *args, **kwargs):
        handler = AddOperationHandler(autodetector, app_label, args, kwargs)

        if isinstance(operation, CreateModel):
            return handler.add_create_model(operation)

        if isinstance(operation, DeleteModel):
            return handler.add_delete_model(operation)

        if isinstance(operation, AddField):
            return handler.add_field(operation)

        if isinstance(operation, RemoveField):
            return handler.remove_field(operation)

        if isinstance(operation, AlterField):
            return handler.alter_field(operation)

        if isinstance(operation, RenameField):
            return handler.rename_field(operation)

        return handler.add(operation)

    with mock.patch(add_operation_path, new=_patched):
        yield



from django.db.models.base import ModelBase

from psqlextra.types import PostgresPartitioningMethod

from .base import PostgresModel
from .options import PostgresPartitionedModelOptions


class PostgresPartitionedModelMeta(ModelBase):
    """Custom meta class for :see:PostgresPartitionedModel.

    This meta class extracts attributes from the inner
    `PartitioningMeta` class and copies it onto a `_partitioning_meta`
    attribute. This is similar to how Django's `_meta` works.
    """

    default_method = PostgresPartitioningMethod.RANGE
    default_key = []

    def __new__(cls, name, bases, attrs, **kwargs):
        new_class = super().__new__(cls, name, bases, attrs, **kwargs)
        meta_class = attrs.pop("PartitioningMeta", None)

        method = getattr(meta_class, "method", None)
        key = getattr(meta_class, "key", None)

        patitioning_meta = PostgresPartitionedModelOptions(
            method=method or cls.default_method, key=key or cls.default_key
        )

        new_class.add_to_class("_partitioning_meta", patitioning_meta)
        return new_class


class PostgresPartitionedModel(
    PostgresModel, metaclass=PostgresPartitionedModelMeta
):
    """Base class for taking advantage of PostgreSQL's 11.x native support for
    table partitioning."""

    class Meta:
        abstract = True
        base_manager_name = "objects"



from django.db import models

from psqlextra.util import postgres_manager

from .fake_model import get_fake_model


def test_manager_context():
    """Tests whether the :see:postgres_manager context manager can be used to
    get access to :see:PostgresManager on a model that does not use it directly
    or inherits from :see:PostgresModel."""

    model = get_fake_model(
        {"myfield": models.CharField(max_length=255, unique=True)}, models.Model
    )

    with postgres_manager(model) as manager:
        manager.upsert(conflict_target=["myfield"], fields=dict(myfield="beer"))

        assert manager.first().myfield == "beer"



import uuid

import pytest

from django.db import models, transaction
from django.db.utils import IntegrityError

from ..fake_model import get_fake_model


@pytest.mark.benchmark()
def test_upsert_traditional(benchmark):
    model = get_fake_model(
        {"field": models.CharField(max_length=255, unique=True)}
    )

    random_value = str(uuid.uuid4())[:8]
    model.objects.create(field=random_value)

    def _traditional_upsert(model, random_value):
        """Performs a concurrency safe upsert the traditional way."""

        try:

            with transaction.atomic():
                return model.objects.create(field=random_value)
        except IntegrityError:
            model.objects.update(field=random_value)
            return model.objects.get(field=random_value)

    benchmark(_traditional_upsert, model, random_value)


@pytest.mark.benchmark()
def test_upsert_native(benchmark):
    model = get_fake_model(
        {"field": models.CharField(max_length=255, unique=True)}
    )

    random_value = str(uuid.uuid4())[:8]
    model.objects.create(field=random_value)

    def _native_upsert(model, random_value):
        """Performs a concurrency safe upsert using the native PostgreSQL
        upsert."""

        return model.objects.upsert_and_get(
            conflict_target=["field"], fields=dict(field=random_value)
        )

    benchmark(_native_upsert, model, random_value)



from datetime import datetime, timezone
from typing import Generator, Optional

from dateutil.relativedelta import relativedelta

from .range_strategy import PostgresRangePartitioningStrategy
from .time_partition import PostgresTimePartition
from .time_partition_size import PostgresTimePartitionSize


class PostgresCurrentTimePartitioningStrategy(
    PostgresRangePartitioningStrategy
):
    """Implments a time based partitioning strategy where each partition
    contains values for a specific time period.

    All buckets will be equal in size and start at the start of the
    unit. With monthly partitioning, partitions start on the 1st and
    with weekly partitioning, partitions start on monday.
    """

    def __init__(
        self,
        size: PostgresTimePartitionSize,
        count: int,
        max_age: Optional[relativedelta] = None,
    ) -> None:
        """Initializes a new instance of :see:PostgresTimePartitioningStrategy.

        Arguments:
            size:
                The size of each partition.

            count:
                The amount of partitions to create ahead
                from the current date/time.

            max_age:
                Maximum age of a partition. Partitions
                older than this are deleted during
                auto cleanup.
        """

        self.size = size
        self.count = count
        self.max_age = max_age

    def to_create(self) -> Generator[PostgresTimePartition, None, None]:
        current_datetime = self.size.start(self.get_start_datetime())

        for _ in range(self.count):
            yield PostgresTimePartition(
                start_datetime=current_datetime, size=self.size
            )

            current_datetime += self.size.as_delta()

    def to_delete(self) -> Generator[PostgresTimePartition, None, None]:
        if not self.max_age:
            return

        current_datetime = self.size.start(
            self.get_start_datetime() - self.max_age
        )

        while True:
            yield PostgresTimePartition(
                start_datetime=current_datetime, size=self.size
            )

            current_datetime -= self.size.as_delta()

    def get_start_datetime(self) -> datetime:
        return datetime.now(timezone.utc)



from contextlib import contextmanager
from unittest import mock

from django.db.migrations.state import ProjectState

from psqlextra.models import (
    PostgresMaterializedViewModel,
    PostgresPartitionedModel,
    PostgresViewModel,
)

from .state import (
    PostgresMaterializedViewModelState,
    PostgresPartitionedModelState,
    PostgresViewModelState,
)

# original `ProjectState.from_apps` function,
# saved here so the patched version can call
# the original
original_from_apps = ProjectState.from_apps


def project_state_from_apps(apps):
    """Creates a :see:ProjectState instance from the specified list of apps."""

    project_state = original_from_apps(apps)
    for model in apps.get_models(include_swapped=True):
        model_state = None

        # for some of our custom models, use the more specific model
        # state.. for everything else, business as usual
        if issubclass(model, PostgresPartitionedModel):
            model_state = PostgresPartitionedModelState.from_model(model)
        elif issubclass(model, PostgresMaterializedViewModel):
            model_state = PostgresMaterializedViewModelState.from_model(model)
        elif issubclass(model, PostgresViewModel):
            model_state = PostgresViewModelState.from_model(model)
        else:
            continue

        model_state_key = (model_state.app_label, model_state.name_lower)
        project_state.models[model_state_key] = model_state

    return project_state


@contextmanager
def patched_project_state():
    """Patches the standard Django :see:ProjectState.from_apps for the duration
    of the context.

    The patch intercepts the `from_apps` function to control
    how model state is creatd. We want to use our custom
    model state classes for certain types of models.

    We have to do this because there is no way in Django
    to extend the project state otherwise.
    """

    from_apps_module_path = "django.db.migrations.state"
    from_apps_class_path = f"{from_apps_module_path}.ProjectState"
    from_apps_path = f"{from_apps_class_path}.from_apps"

    with mock.patch(from_apps_path, new=project_state_from_apps):
        yield



import pytest

from psqlextra.fields import HStoreField


def test_hstore_field_deconstruct():
    """Tests whether the :see:HStoreField's deconstruct() method works
    properly."""

    original_kwargs = dict(uniqueness=["beer", "other"], required=[])
    _, _, _, new_kwargs = HStoreField(**original_kwargs).deconstruct()

    for key, value in original_kwargs.items():
        assert new_kwargs[key] == value


@pytest.mark.parametrize(
    "input,output",
    [
        (dict(key1=1, key2=2), dict(key1="1", key2="2")),
        (dict(key1="1", key2="2"), dict(key1="1", key2="2")),
        (
            dict(key1=1, key2=None, key3="3"),
            dict(key1="1", key2=None, key3="3"),
        ),
        ([1, 2, 3], ["1", "2", "3"]),
        (["1", "2", "3"], ["1", "2", "3"]),
    ],
)
def test_hstore_field_get_prep_value(input, output):
    """Tests whether the :see:HStoreField's get_prep_value method works
    properly."""

    assert HStoreField().get_prep_value(input) == output



from psqlextra.backend.migrations.state import PostgresListPartitionState

from .partition import PostgresPartitionOperation


class PostgresAddListPartition(PostgresPartitionOperation):
    """Adds a new list partition to a :see:PartitionedPostgresModel."""

    def __init__(self, model_name, name, values):
        """Initializes new instance of :see:AddListPartition.

        Arguments:
            model_name:
                The name of the :see:PartitionedPostgresModel.

            name:
                The name to give to the new partition table.

            values:
                Partition key values that should be
                stored in this partition.
        """

        super().__init__(model_name, name)

        self.values = values

    def state_forwards(self, app_label, state):
        model = state.models[(app_label, self.model_name_lower)]
        model.add_partition(
            PostgresListPartitionState(
                app_label=app_label,
                model_name=self.model_name,
                name=self.name,
                values=self.values,
            )
        )

        state.reload_model(app_label, self.model_name_lower)

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        model = to_state.apps.get_model(app_label, self.model_name)
        if self.allow_migrate_model(schema_editor.connection.alias, model):
            schema_editor.add_list_partition(model, self.name, self.values)

    def deconstruct(self):
        name, args, kwargs = super().deconstruct()
        kwargs["values"] = self.values

        return name, args, kwargs

    def describe(self) -> str:
        return "Creates list partition %s on %s" % (self.name, self.model_name)



from django.db import models

from psqlextra.query import ConflictAction

from .fake_model import get_fake_model


def test_insert():
    """Tests whether inserts works when the primary key is explicitly
    specified."""

    model = get_fake_model(
        {"cookies": models.CharField(max_length=255, null=True)}
    )

    pk = model.objects.all().insert(cookies="some-cookies")

    assert pk is not None

    obj1 = model.objects.get()
    assert obj1.pk == pk
    assert obj1.cookies == "some-cookies"


def test_insert_explicit_pk():
    """Tests whether inserts works when the primary key is explicitly
    specified."""

    model = get_fake_model(
        {
            "name": models.CharField(max_length=255, primary_key=True),
            "cookies": models.CharField(max_length=255, null=True),
        }
    )

    pk = model.objects.all().insert(name="the-object", cookies="some-cookies")

    assert pk == "the-object"

    obj1 = model.objects.get()
    assert obj1.pk == "the-object"
    assert obj1.name == "the-object"
    assert obj1.cookies == "some-cookies"


def test_insert_on_conflict():
    """Tests whether inserts works when a conflict is anticipated."""

    model = get_fake_model(
        {
            "name": models.CharField(max_length=255, unique=True),
            "cookies": models.CharField(max_length=255, null=True),
        }
    )

    pk = model.objects.on_conflict([("pk")], ConflictAction.NOTHING).insert(
        name="the-object", cookies="some-cookies"
    )

    assert pk is not None

    obj1 = model.objects.get()
    assert obj1.pk == pk
    assert obj1.name == "the-object"
    assert obj1.cookies == "some-cookies"


def test_insert_on_conflict_explicit_pk():
    """Tests whether inserts works when a conflict is anticipated and the
    primary key is explicitly specified."""

    model = get_fake_model(
        {
            "name": models.CharField(max_length=255, primary_key=True),
            "cookies": models.CharField(max_length=255, null=True),
        }
    )

    pk = model.objects.on_conflict([("name")], ConflictAction.NOTHING).insert(
        name="the-object", cookies="some-cookies"
    )

    assert pk == "the-object"

    obj1 = model.objects.get()
    assert obj1.pk == "the-object"
    assert obj1.name == "the-object"
    assert obj1.cookies == "some-cookies"


def test_insert_with_different_column_name():
    """Tests whether inserts works when the primary key is explicitly
    specified."""

    model = get_fake_model(
        {
            "name": models.CharField(max_length=255, primary_key=True),
            "cookies": models.CharField(
                max_length=255, null=True, db_column="brownies"
            ),
        }
    )

    cookie_string = "these-are-brownies"

    results = model.objects.on_conflict(
        ["name"], ConflictAction.NOTHING
    ).insert_and_get(name="the-object", cookies=cookie_string)

    assert results is not None
    assert results.cookies == cookie_string

    obj1 = model.objects.get()
    assert obj1.cookies == cookie_string


def test_insert_many_to_many():
    """Tests whether adding a rows to a m2m works after using insert_and_get.

    The model returned by `insert_and_get` must be configured in a
    special way. Just creating a instance of the model is not enough to
    be able to add m2m rows.
    """

    model1 = get_fake_model({"name": models.TextField(primary_key=True)})

    model2 = get_fake_model(
        {
            "name": models.TextField(primary_key=True),
            "model1s": models.ManyToManyField(model1),
        }
    )

    row2 = model2.objects.on_conflict(
        ["name"], ConflictAction.UPDATE
    ).insert_and_get(name="swen")

    row1 = model1.objects.create(name="booh")

    row2.model1s.add(row1)
    row2.save()



from typing import Optional

from dateutil.relativedelta import relativedelta

from psqlextra.models import PostgresPartitionedModel

from .config import PostgresPartitioningConfig
from .current_time_strategy import PostgresCurrentTimePartitioningStrategy
from .time_partition_size import PostgresTimePartitionSize


def partition_by_current_time(
    model: PostgresPartitionedModel,
    count: int,
    years: Optional[int] = None,
    months: Optional[int] = None,
    weeks: Optional[int] = None,
    days: Optional[int] = None,
    max_age: Optional[relativedelta] = None,
) -> PostgresPartitioningConfig:
    """Short-hand for generating a partitioning config that partitions the
    specified model by time.

    One specifies one of the `years`, `months`, `weeks`
    or `days` parameter to indicate the size of each
    partition. These parameters cannot be combined.

    Arguments:
        count:
            The amount of partitions to create ahead of
            the current date/time.

        years:
            The amount of years each partition should contain.

        months:
            The amount of months each partition should contain.

        weeks:
            The amount of weeks each partition should contain.

        days:
            The amount of days each partition should contain.

        max_age:
            The maximum age of a partition (calculated from the
            start of the partition).

            Partitions older than this are deleted when running
            a delete/cleanup run.
    """

    size = PostgresTimePartitionSize(
        years=years, months=months, weeks=weeks, days=days
    )

    return PostgresPartitioningConfig(
        model=model,
        strategy=PostgresCurrentTimePartitioningStrategy(
            size=size, count=count, max_age=max_age
        ),
    )


__all_ = ["partition_by_current_time"]



import pytest

from django.contrib.postgres.signals import register_type_handlers
from django.db import connection

from .fake_model import define_fake_app


@pytest.fixture(scope="function", autouse=True)
def database_access(db):
    """Automatically enable database access for all tests."""

    # enable the hstore extension on our database because
    # our tests rely on it...
    with connection.schema_editor() as schema_editor:
        schema_editor.execute("CREATE EXTENSION IF NOT EXISTS hstore")
        register_type_handlers(schema_editor.connection)


@pytest.fixture
def fake_app():
    """Creates a fake Django app and deletes it at the end of the test."""

    with define_fake_app() as fake_app:
        yield fake_app



from dataclasses import dataclass
from typing import List, Optional

from psqlextra.types import PostgresPartitioningMethod

from . import base_impl


@dataclass
class PostgresIntrospectedPartitionTable:
    """Data container for information about a partition."""

    name: str
    full_name: str
    comment: Optional[str]


@dataclass
class PostgresIntrospectedPartitonedTable:
    """Data container for information about a partitioned table."""

    name: str
    method: PostgresPartitioningMethod
    key: List[str]
    partitions: List[PostgresIntrospectedPartitionTable]

    def partition_by_name(
        self, name: str
    ) -> Optional[PostgresIntrospectedPartitionTable]:
        """Finds the partition with the specified name."""

        return next(
            (
                partition
                for partition in self.partitions
                if partition.name == name
            ),
            None,
        )


class PostgresIntrospection(base_impl.introspection()):
    """Adds introspection features specific to PostgreSQL."""

    def get_partitioned_tables(
        self, cursor
    ) -> PostgresIntrospectedPartitonedTable:
        """Gets a list of partitioned tables."""

        sql = """
            SELECT
                pg_class.relname,
                pg_partitioned_table.partstrat
            FROM
                pg_partitioned_table
            JOIN
                pg_class
            ON
                pg_class.oid = pg_partitioned_table.partrelid
        """

        cursor.execute(sql)

        return [
            PostgresIntrospectedPartitonedTable(
                name=row[0],
                method=PostgresPartitioningMethod.RANGE
                if row[1] == "r"
                else PostgresPartitioningMethod.LIST,
                key=self.get_partition_key(cursor, row[0]),
                partitions=self.get_partitions(cursor, row[0]),
            )
            for row in cursor.fetchall()
        ]

    def get_partitioned_table(self, cursor, table_name: str):
        """Gets a single partitioned table."""

        return next(
            (
                table
                for table in self.get_partitioned_tables(cursor)
                if table.name == table_name
            ),
            None,
        )

    def get_partitions(
        self, cursor, table_name
    ) -> List[PostgresIntrospectedPartitionTable]:
        """Gets a list of partitions belonging to the specified partitioned
        table."""

        sql = """
            SELECT
                child.relname,
                pg_description.description
            FROM pg_inherits
            JOIN
                pg_class parent
            ON
                pg_inherits.inhparent = parent.oid
            JOIN
                pg_class child
            ON
                pg_inherits.inhrelid = child.oid
            JOIN
                pg_namespace nmsp_parent
            ON
                nmsp_parent.oid = parent.relnamespace
            JOIN
                pg_namespace nmsp_child
            ON
                nmsp_child.oid = child.relnamespace
            LEFT JOIN
                pg_description
            ON
                pg_description.objoid = child.oid
            WHERE
                parent.relname = %s
        """

        cursor.execute(sql, (table_name,))

        return [
            PostgresIntrospectedPartitionTable(
                name=row[0].replace(f"{table_name}_", ""),
                full_name=row[0],
                comment=row[1] or None,
            )
            for row in cursor.fetchall()
        ]

    def get_partition_key(self, cursor, table_name: str) -> List[str]:
        """Gets the partition key for the specified partitioned table.

        Returns:
            A list of column names that are part of the
            partition key.
        """

        sql = """
            SELECT
                col.column_name
            FROM
                (SELECT partrelid,
                        partnatts,
                        CASE partstrat
                            WHEN 'l' THEN 'list'
                            WHEN 'r' THEN 'range'
                        END AS partition_strategy,
                        Unnest(partattrs) column_index
                 FROM pg_partitioned_table) pt
            JOIN
                pg_class par
            ON par.oid = pt.partrelid
            JOIN
                information_schema.COLUMNS col
            ON
                col.table_schema = par.relnamespace :: regnamespace :: text
                AND col.table_name = par.relname
                AND ordinal_position = pt.column_index
            WHERE
                table_name = %s
        """

        cursor.execute(sql, (table_name,))
        return [row[0] for row in cursor.fetchall()]

    def get_constraints(self, cursor, table_name: str):
        """Retrieve any constraints or keys (unique, pk, fk, check, index)
        across one or more columns.

        Also retrieve the definition of expression-based indexes.
        """

        constraints = super().get_constraints(cursor, table_name)

        # standard Django implementation does not return the definition
        # for indexes, only for constraints, let's patch that up
        cursor.execute(
            "SELECT indexname, indexdef FROM pg_indexes WHERE tablename = %s",
            (table_name,),
        )
        for index, definition in cursor.fetchall():
            if constraints[index].get("definition") is None:
                constraints[index]["definition"] = definition

        return constraints



import pytest

from django.db import IntegrityError, connection, models
from django.db.migrations import AddIndex, CreateModel

from psqlextra.indexes import CaseInsensitiveUniqueIndex
from psqlextra.models import PostgresModel

from .fake_model import get_fake_model
from .migrations import apply_migration, filtered_schema_editor


def test_ciui_migrations():
    """Tests whether migrations for case sensitive indexes are being created as
    expected."""

    index_1 = CaseInsensitiveUniqueIndex(
        fields=["name", "other_name"], name="index1"
    )

    ops = [
        CreateModel(
            name="mymodel",
            fields=[
                ("name", models.CharField(max_length=255)),
                ("other_name", models.CharField(max_length=255)),
            ],
        ),
        AddIndex(model_name="mymodel", index=index_1),
    ]

    with filtered_schema_editor("CREATE UNIQUE INDEX") as calls:
        apply_migration(ops)

    sql = str([call[0] for _, call, _ in calls["CREATE UNIQUE INDEX"]][0])
    expected_sql = 'CREATE UNIQUE INDEX "index1" ON "tests_mymodel" (LOWER("name"), LOWER("other_name"))'
    assert sql == expected_sql


def test_ciui():
    """Tests whether the case insensitive unique index works as expected."""

    index_1 = CaseInsensitiveUniqueIndex(fields=["name"], name="index1")

    model = get_fake_model(
        {"name": models.CharField(max_length=255)}, PostgresModel
    )

    with connection.schema_editor() as schema_editor:
        schema_editor.add_index(model, index_1)

    model.objects.create(name="henk")

    with pytest.raises(IntegrityError):
        model.objects.create(name="Henk")


def test_ciui_on_conflict():
    """Tests wether fields with a :see:CaseInsensitiveUniqueIndex can be used
    as a conflict target."""

    index_1 = CaseInsensitiveUniqueIndex(fields=["name"], name="index1")

    model = get_fake_model(
        {"name": models.CharField(max_length=255)},
        PostgresModel,
        {"indexes": [index_1]},
    )

    model.objects.upsert(conflict_target=["name"], fields=dict(name="henk"))



from typing import Dict, List, Optional, Union

from psqlextra.types import PostgresPartitioningMethod, SQLWithParams


class PostgresPartitionedModelOptions:
    """Container for :see:PostgresPartitionedModel options.

    This is where attributes copied from the model's `PartitioningMeta`
    are held.
    """

    def __init__(self, method: PostgresPartitioningMethod, key: List[str]):
        self.method = method
        self.key = key
        self.original_attrs: Dict[
            str, Union[PostgresPartitioningMethod, List[key]]
        ] = dict(method=method, key=key)


class PostgresViewOptions:
    """Container for :see:PostgresView and :see:PostgresMaterializedView
    options.

    This is where attributes copied from the model's `ViewMeta` are
    held.
    """

    def __init__(self, query: Optional[SQLWithParams]):
        self.query = query
        self.original_attrs: Dict[str, Optional[SQLWithParams]] = dict(
            query=self.query
        )



from django.db.migrations.operations.base import Operation


class ApplyState(Operation):
    """Takes an abritrary operation and migrates the project state but does not
    apply the operation to the database.

    This is very similar to the :see:RunSQL `state_operations`
    parameter. This is useful if you want to tell Django that an
    operation was applied without actually applying it.
    """

    reduces_to_sql = False

    def __init__(self, state_operation: Operation) -> None:
        self.state_operation = state_operation

    def deconstruct(self):
        kwargs = {"state_operation": self.state_operation}
        return (self.__class__.__qualname__, [], kwargs)

    @property
    def reversible(self):
        return True

    def state_forwards(self, app_label, state):
        self.state_operation.state_forwards(app_label, state)

    def state_backwards(self, app_label, state):
        self.state_operation.state_backwards(app_label, state)

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        pass

    def database_backwards(
        self, app_label, schema_editor, from_state, to_state
    ):
        pass

    def describe(self):
        return "Apply state: " + self.state_operation.describe()



from django.db import connection, models

from psqlextra.backend.schema import PostgresSchemaEditor

from .fake_model import define_fake_materialized_view_model, get_fake_model


def test_materialized_view_model_refresh():
    """Tests whether a materialized view can be refreshed."""

    underlying_model = get_fake_model({"name": models.TextField()})

    model = define_fake_materialized_view_model(
        {"name": models.TextField()},
        {"query": underlying_model.objects.filter(name="test1")},
    )

    underlying_model.objects.create(name="test1")
    underlying_model.objects.create(name="test2")

    schema_editor = PostgresSchemaEditor(connection)
    schema_editor.create_materialized_view_model(model)

    # materialized view should only show records name="test"1
    objs = list(model.objects.all())
    assert len(objs) == 1
    assert objs[0].name == "test1"

    # create another record with "test1" and refresh
    underlying_model.objects.create(name="test1")
    model.refresh()

    objs = list(model.objects.all())
    assert len(objs) == 2



import uuid

import pytest

from django.apps import apps
from django.db import models

from psqlextra.backend.migrations.state import (
    PostgresPartitionedModelState,
    PostgresPartitionState,
)
from psqlextra.manager import PostgresManager
from psqlextra.models import PostgresPartitionedModel
from psqlextra.types import PostgresPartitioningMethod

from .fake_model import define_fake_partitioned_model


@pytest.fixture
def model():
    fields = {"name": models.TextField(), "category": models.TextField()}

    partitioning_options = {
        "method": PostgresPartitioningMethod.LIST,
        "key": ["category"],
    }

    model = define_fake_partitioned_model(fields, partitioning_options)
    return model


def test_partitioned_model_state_copies():
    """Tests whether cloning the model state properly copies all the options.

    If it does not copy them, bad things can happen as the state is
    mutated to build up migration state.
    """

    options = dict(method=PostgresPartitioningMethod.RANGE, key=["timestamp"])

    state = PostgresPartitionedModelState(
        app_label="tests",
        name=str(uuid.uuid4()),
        fields=[],
        options=None,
        partitioning_options=options,
        bases=(PostgresPartitionedModel,),
    )

    assert options is not state.partitioning_options


def test_partitioned_model_state_from_model(model):
    """Tests whether creating state from an existing model works as
    expected."""

    state = PostgresPartitionedModelState.from_model(model)
    assert state.partitions == {}
    assert (
        state.partitioning_options["method"] == model._partitioning_meta.method
    )
    assert state.partitioning_options["key"] == model._partitioning_meta.key


def test_partitioned_model_clone(model):
    """Tests whether cloning the state actually clones the partitioning
    options.

    If its not a copy, but a reference instead, bad things can happen as
    the options are mutated to build up migration state.
    """

    state = PostgresPartitionedModelState.from_model(model)
    state.partitions = {
        "pt1": PostgresPartitionState(
            app_label="tests", model_name="tests", name="pt1"
        )
    }

    state_copy = state.clone()
    assert state.partitions is not state_copy.partitions
    assert state.partitioning_options is not state_copy.partitioning_options


def test_partitioned_model_render(model):
    """Tests whether the state can be rendered into a valid model class."""

    options = dict(method=PostgresPartitioningMethod.RANGE, key=["timestamp"])

    state = PostgresPartitionedModelState(
        app_label="tests",
        name=str(uuid.uuid4()),
        fields=[("name", models.TextField())],
        options=None,
        partitioning_options=options,
        bases=(PostgresPartitionedModel,),
        managers=[("cookie", PostgresManager())],
    )

    rendered_model = state.render(apps)

    assert issubclass(rendered_model, PostgresPartitionedModel)
    assert rendered_model.name
    assert isinstance(rendered_model.objects, PostgresManager)
    assert isinstance(rendered_model.cookie, PostgresManager)
    assert rendered_model.__name__ == state.name
    assert rendered_model._meta.apps == apps
    assert rendered_model._meta.app_label == "tests"
    assert rendered_model._partitioning_meta.method == options["method"]
    assert rendered_model._partitioning_meta.key == options["key"]



from psqlextra.backend.migrations.state import PostgresPartitionState

from .partition import PostgresPartitionOperation


class PostgresAddDefaultPartition(PostgresPartitionOperation):
    """Adds a new default partition to a :see:PartitionedPostgresModel."""

    def state_forwards(self, app_label, state):
        model_state = state.models[(app_label, self.model_name_lower)]
        model_state.add_partition(
            PostgresPartitionState(
                app_label=app_label, model_name=self.model_name, name=self.name
            )
        )

        state.reload_model(app_label, self.model_name_lower)

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        model = to_state.apps.get_model(app_label, self.model_name)
        if self.allow_migrate_model(schema_editor.connection.alias, model):
            schema_editor.add_default_partition(model, self.name)

    def database_backwards(
        self, app_label, schema_editor, from_state, to_state
    ):
        model = from_state.apps.get_model(app_label, self.model_name)
        if self.allow_migrate_model(schema_editor.connection.alias, model):
            schema_editor.delete_partition(model, self.name)

    def describe(self) -> str:
        return "Creates default partition '%s' on %s" % (
            self.name,
            self.model_name,
        )



from django.db import connection, models

from psqlextra.backend.schema import PostgresSchemaEditor

from . import db_introspection
from .fake_model import (
    define_fake_materialized_view_model,
    define_fake_view_model,
    get_fake_model,
)


def test_schema_editor_create_delete_view():
    """Tests whether creating and then deleting a view using the schema editor
    works as expected."""

    underlying_model = get_fake_model({"name": models.TextField()})

    model = define_fake_view_model(
        {"name": models.TextField()},
        {"query": underlying_model.objects.filter(name="test1")},
    )

    underlying_model.objects.create(name="test1")
    underlying_model.objects.create(name="test2")

    schema_editor = PostgresSchemaEditor(connection)
    schema_editor.create_view_model(model)

    # view should only show records name="test"1
    objs = list(model.objects.all())
    assert len(objs) == 1
    assert objs[0].name == "test1"

    # create another record, view should have it right away
    underlying_model.objects.create(name="test1")
    assert model.objects.count() == 2

    # delete the view
    schema_editor.delete_view_model(model)

    # make sure it was actually deleted
    assert model._meta.db_table not in db_introspection.table_names(True)


def test_schema_editor_replace_view():
    """Tests whether creating a view and then replacing it with another one
    (thus changing the backing query) works as expected."""

    underlying_model = get_fake_model({"name": models.TextField()})

    model = define_fake_view_model(
        {"name": models.TextField()},
        {"query": underlying_model.objects.filter(name="test1")},
    )

    underlying_model.objects.create(name="test1")
    underlying_model.objects.create(name="test2")

    schema_editor = PostgresSchemaEditor(connection)
    schema_editor.create_view_model(model)

    objs = list(model.objects.all())
    assert len(objs) == 1
    assert objs[0].name == "test1"

    model._view_meta.query = underlying_model.objects.filter(
        name="test2"
    ).query.sql_with_params()
    schema_editor.replace_view_model(model)

    objs = list(model.objects.all())
    assert len(objs) == 1
    assert objs[0].name == "test2"


def test_schema_editor_create_delete_materialized_view():
    """Tests whether creating and then deleting a materialized view using the
    schema editor works as expected."""

    underlying_model = get_fake_model({"name": models.TextField()})

    model = define_fake_materialized_view_model(
        {"name": models.TextField()},
        {"query": underlying_model.objects.filter(name="test1")},
    )

    underlying_model.objects.create(name="test1")
    underlying_model.objects.create(name="test2")

    schema_editor = PostgresSchemaEditor(connection)
    schema_editor.create_materialized_view_model(model)

    # materialized view should only show records name="test"1
    objs = list(model.objects.all())
    assert len(objs) == 1
    assert objs[0].name == "test1"

    # delete the materialized view
    schema_editor.delete_materialized_view_model(model)

    # make sure it was actually deleted
    assert model._meta.db_table not in db_introspection.table_names(True)


def test_schema_editor_replace_materialized_view():
    """Tests whether creating a materialized view and then replacing it with
    another one (thus changing the backing query) works as expected."""

    underlying_model = get_fake_model({"name": models.TextField()})

    model = define_fake_materialized_view_model(
        {"name": models.TextField()},
        {"query": underlying_model.objects.filter(name="test1")},
        {"indexes": [models.Index(fields=["name"])]},
    )

    underlying_model.objects.create(name="test1")
    underlying_model.objects.create(name="test2")

    schema_editor = PostgresSchemaEditor(connection)
    schema_editor.create_materialized_view_model(model)

    for index in model._meta.indexes:
        schema_editor.add_index(model, index)

    constraints_before = db_introspection.get_constraints(model._meta.db_table)

    objs = list(model.objects.all())
    assert len(objs) == 1
    assert objs[0].name == "test1"

    model._view_meta.query = underlying_model.objects.filter(
        name="test2"
    ).query.sql_with_params()
    schema_editor.replace_materialized_view_model(model)

    objs = list(model.objects.all())
    assert len(objs) == 1
    assert objs[0].name == "test2"

    # make sure all indexes/constraints still exists because
    # replacing a materialized view involves re-creating it
    constraints_after = db_introspection.get_constraints(model._meta.db_table)
    assert constraints_after == constraints_before



from datetime import datetime
from typing import Optional

from dateutil.relativedelta import relativedelta

from .current_time_strategy import PostgresCurrentTimePartitioningStrategy
from .time_partition_size import PostgresTimePartitionSize


class PostgresTimePartitioningStrategy(PostgresCurrentTimePartitioningStrategy):
    def __init__(
        self,
        start_datetime: datetime,
        size: PostgresTimePartitionSize,
        count: int,
        max_age: Optional[relativedelta] = None,
    ) -> None:
        super().__init__(size, count, max_age)

        self.start_datetime = start_datetime

    def get_start_datetime(self) -> datetime:
        return self.start_datetime



class PostgresPartitioningError(RuntimeError):
    """Raised when the partitioning configuration is broken or automatically
    creating/deleting partitions fails."""


__all__ = ["PostgresPartitioningError"]



from typing import Dict, List, Type

from psqlextra.models import PostgresPartitionedModel

from .model import PostgresModelState


class PostgresPartitionState:
    """Represents the state of a partition for a :see:PostgresPartitionedModel
    during a migration."""

    def __init__(self, app_label: str, model_name: str, name: str) -> None:
        self.app_label = app_label
        self.model_name = model_name
        self.name = name


class PostgresRangePartitionState(PostgresPartitionState):
    """Represents the state of a range partition for a
    :see:PostgresPartitionedModel during a migration."""

    def __init__(
        self, app_label: str, model_name: str, name: str, from_values, to_values
    ):
        super().__init__(app_label, model_name, name)

        self.from_values = from_values
        self.to_values = to_values


class PostgresListPartitionState(PostgresPartitionState):
    """Represents the state of a list partition for a
    :see:PostgresPartitionedModel during a migration."""

    def __init__(self, app_label: str, model_name: str, name: str, values):
        super().__init__(app_label, model_name, name)

        self.values = values


class PostgresPartitionedModelState(PostgresModelState):
    """Represents the state of a :see:PostgresPartitionedModel in the
    migrations."""

    def __init__(
        self,
        *args,
        partitions: List[PostgresPartitionState] = [],
        partitioning_options={},
        **kwargs
    ):
        """Initializes a new instance of :see:PostgresPartitionedModelState.

        Arguments:
            partitioning_options:
                Dictionary of options for partitioning.

                See: PostgresPartitionedModelMeta for a list.
        """

        super().__init__(*args, **kwargs)

        self.partitions: Dict[str, PostgresPartitionState] = {
            partition.name: partition for partition in partitions
        }
        self.partitioning_options = dict(partitioning_options)

    def add_partition(self, partition: PostgresPartitionState):
        """Adds a partition to this partitioned model state."""

        self.partitions[partition.name] = partition

    def delete_partition(self, name: str):
        """Deletes a partition from this partitioned model state."""

        del self.partitions[name]

    @classmethod
    def _pre_new(
        cls,
        model: PostgresPartitionedModel,
        model_state: "PostgresPartitionedModelState",
    ) -> "PostgresPartitionedModelState":
        """Called when a new model state is created from the specified
        model."""

        model_state.partitions = dict()
        model_state.partitioning_options = dict(
            model._partitioning_meta.original_attrs
        )
        return model_state

    def _pre_clone(
        self, model_state: "PostgresPartitionedModelState"
    ) -> "PostgresPartitionedModelState":
        """Called when this model state is cloned."""

        model_state.partitions = dict(self.partitions)
        model_state.partitioning_options = dict(self.partitioning_options)
        return model_state

    def _pre_render(self, name: str, bases, attributes):
        """Called when this model state is rendered into a model."""

        partitioning_meta = type(
            "PartitioningMeta", (), dict(self.partitioning_options)
        )
        return (
            name,
            bases,
            {**attributes, "PartitioningMeta": partitioning_meta},
        )

    @classmethod
    def _get_base_model_class(self) -> Type[PostgresPartitionedModel]:
        """Gets the class to use as a base class for rendered models."""

        return PostgresPartitionedModel



from abc import abstractmethod
from typing import Optional

from psqlextra.backend.schema import PostgresSchemaEditor
from psqlextra.models import PostgresPartitionedModel


class PostgresPartition:
    """Base class for a PostgreSQL table partition."""

    @abstractmethod
    def name(self) -> str:
        """Generates/computes the name for this partition."""

    @abstractmethod
    def create(
        self,
        model: PostgresPartitionedModel,
        schema_editor: PostgresSchemaEditor,
        comment: Optional[str] = None,
    ) -> None:
        """Creates this partition in the database."""

    @abstractmethod
    def delete(
        self,
        model: PostgresPartitionedModel,
        schema_editor: PostgresSchemaEditor,
    ) -> None:
        """Deletes this partition from the database."""

    def deconstruct(self) -> dict:
        """Deconstructs this partition into a dict of attributes/fields."""

        return {"name": self.name()}


__all__ = ["PostgresPartition"]



import pytest

from django.core.exceptions import ImproperlyConfigured
from django.db import models

from psqlextra.models import PostgresMaterializedViewModel, PostgresViewModel

from .fake_model import define_fake_model, define_fake_view_model


@pytest.mark.parametrize(
    "model_base", [PostgresViewModel, PostgresMaterializedViewModel]
)
def test_view_model_meta_query_set(model_base):
    """Tests whether you can set a :see:QuerySet to be used as the underlying
    query for a view."""

    model = define_fake_model({"name": models.TextField()})

    view_model = define_fake_view_model(
        {"name": models.TextField()},
        model_base=model_base,
        view_options={"query": model.objects.all()},
    )

    expected_sql = 'SELECT "{0}"."id", "{0}"."name" FROM "{0}"'.format(
        model._meta.db_table
    )
    assert view_model._view_meta.query == (expected_sql, tuple())


@pytest.mark.parametrize(
    "model_base", [PostgresViewModel, PostgresMaterializedViewModel]
)
@pytest.mark.parametrize("bind_params", [("test",), ["test"]])
def test_view_model_meta_sql_with_params(model_base, bind_params):
    """Tests whether you can set a raw SQL query with a tuple of bind params as
    the underlying query for a view."""

    model = define_fake_model({"name": models.TextField()})
    sql = "select * from %s where name = %s" % (model._meta.db_table, "%s")
    sql_with_params = (sql, bind_params)

    view_model = define_fake_view_model(
        {"name": models.TextField()},
        model_base=model_base,
        view_options={"query": sql_with_params},
    )

    assert view_model._view_meta.query == sql_with_params


@pytest.mark.parametrize(
    "model_base", [PostgresViewModel, PostgresMaterializedViewModel]
)
def test_view_model_meta_sql_with_named_params(model_base):
    """Tests whether you can set a raw SQL query with a tuple of bind params as
    the underlying query for a view."""

    model = define_fake_model({"name": models.TextField()})
    sql = "select * from " + model._meta.db_table + " where name = %(name)s"
    sql_with_params = (sql, dict(name="test"))

    view_model = define_fake_view_model(
        {"name": models.TextField()},
        model_base=model_base,
        view_options={"query": sql_with_params},
    )

    assert view_model._view_meta.query == sql_with_params


@pytest.mark.parametrize(
    "model_base", [PostgresViewModel, PostgresMaterializedViewModel]
)
def test_view_model_meta_sql(model_base):
    """Tests whether you can set a raw SQL query without any params."""

    sql = "select 1"

    view_model = define_fake_view_model(
        {"name": models.TextField()},
        model_base=model_base,
        view_options={"query": sql},
    )

    assert view_model._view_meta.query == (sql, tuple())


@pytest.mark.parametrize(
    "model_base", [PostgresViewModel, PostgresMaterializedViewModel]
)
@pytest.mark.parametrize(
    "view_query",
    [
        dict(a=1),
        tuple("test"),
        ("test", None),
        (None, None),
        (1, 2),
        ("select 1", ("a", "b"), "onetoomay"),
    ],
)
def test_view_model_meta_bad_query(model_base, view_query):
    """Tests whether a bad view query configuration raises and error."""

    with pytest.raises(ImproperlyConfigured):
        define_fake_view_model(
            {"name": models.TextField()},
            model_base=model_base,
            view_options={"query": view_query},
        )



from collections import OrderedDict
from itertools import chain
from typing import Dict, Iterable, List, Optional, Tuple, Union

from django.core.exceptions import SuspiciousOperation
from django.db import models, router
from django.db.models.fields import NOT_PROVIDED

from .sql import PostgresInsertQuery, PostgresQuery
from .types import ConflictAction

ConflictTarget = List[Union[str, Tuple[str]]]


class PostgresQuerySet(models.QuerySet):
    """Adds support for PostgreSQL specifics."""

    def __init__(self, model=None, query=None, using=None, hints=None):
        """Initializes a new instance of :see:PostgresQuerySet."""

        super().__init__(model, query, using, hints)

        self.query = query or PostgresQuery(self.model)

        self.conflict_target = None
        self.conflict_action = None
        self.index_predicate = None

    def annotate(self, **annotations):
        """Custom version of the standard annotate function that allows using
        field names as annotated fields.

        Normally, the annotate function doesn't allow you to use the
        name of an existing field on the model as the alias name. This
        version of the function does allow that.

        This is done by temporarily renaming the fields in order to avoid the
        check for conflicts that the base class does.
        We rename all fields instead of the ones that already exist because
        the annotations are stored in an OrderedDict. Renaming only the
        conflicts will mess up the order.
        """
        new_annotations = OrderedDict()
        renames = {}
        for name, value in annotations.items():
            new_name = "%s_new" % name
            new_annotations[new_name] = value
            renames[new_name] = name

        # run the base class's annotate function
        result = super().annotate(**new_annotations)

        # rename the annotations back to as specified
        result.rename_annotations(**renames)
        return result

    def rename_annotations(self, **annotations):
        """Renames the aliases for the specified annotations:

            .annotate(myfield=F('somestuf__myfield'))
            .rename_annotations(myfield='field')

        Arguments:
            annotations:
                The annotations to rename. Mapping the
                old name to the new name.
        """

        self.query.rename_annotations(annotations)
        return self

    def on_conflict(
        self,
        fields: ConflictTarget,
        action: ConflictAction,
        index_predicate: Optional[str] = None,
    ):
        """Sets the action to take when conflicts arise when attempting to
        insert/create a new row.

        Arguments:
            fields:
                The fields the conflicts can occur in.

            action:
                The action to take when the conflict occurs.

            index_predicate:
                The index predicate to satisfy an arbiter partial index (i.e. what partial index to use for checking
                conflicts)
        """

        self.conflict_target = fields
        self.conflict_action = action
        self.index_predicate = index_predicate

        return self

    def bulk_insert(
        self,
        rows: List[dict],
        return_model: bool = False,
        using: Optional[str] = None,
    ):
        """Creates multiple new records in the database.

        This allows specifying custom conflict behavior using .on_conflict().
        If no special behavior was specified, this uses the normal Django create(..)

        Arguments:
            rows:
                An iterable of dictionaries, where each dictionary
                describes the fields to insert.

            return_model (default: False):
                If model instances should be returned rather than
                just dicts.

            using:
                Name of the database connection to use for
                this query.

        Returns:
            A list of either the dicts of the rows inserted, including the pk or
            the models of the rows inserted with defaults for any fields not specified
        """

        if not self.conflict_target and not self.conflict_action:
            # no special action required, use the standard Django bulk_create(..)
            return super().bulk_create(
                [self.model(**fields) for fields in rows]
            )

        deduped_rows = rows

        # when we do a ConflictAction.NOTHING, we are actually
        # doing a ON CONFLICT DO UPDATE with a trick to avoid
        # touching conflicting rows... however, ON CONFLICT UPDATE
        # barfs when you specify the exact same row twice:
        #
        # > "cannot affect row a second time"
        #
        # we filter out the duplicates here to make sure we maintain
        # the same behaviour as the real ON CONFLICT DO NOTHING
        if self.conflict_action == ConflictAction.NOTHING:
            deduped_rows = []
            for row in rows:
                if row in deduped_rows:
                    continue

                deduped_rows.append(row)

        compiler = self._build_insert_compiler(deduped_rows, using=using)
        objs = compiler.execute_sql(return_id=True)
        if return_model:
            return [
                self._create_model_instance(dict(row, **obj), compiler.using)
                for row, obj in zip(deduped_rows, objs)
            ]
        else:
            return [dict(row, **obj) for row, obj in zip(deduped_rows, objs)]

    def insert(self, using: Optional[str] = None, **fields):
        """Creates a new record in the database.

        This allows specifying custom conflict behavior using .on_conflict().
        If no special behavior was specified, this uses the normal Django create(..)

        Arguments:
            fields:
                The fields of the row to create.

            using:
                The name of the database connection
                to use for this query.

        Returns:
            The primary key of the record that was created.
        """

        if self.conflict_target or self.conflict_action:
            compiler = self._build_insert_compiler([fields], using=using)
            rows = compiler.execute_sql(return_id=True)

            pk_field_name = self.model._meta.pk.name
            if not rows or len(rows) == 0:
                return None

            return rows[0][pk_field_name]

        # no special action required, use the standard Django create(..)
        return super().create(**fields).pk

    def insert_and_get(self, using: Optional[str] = None, **fields):
        """Creates a new record in the database and then gets the entire row.

        This allows specifying custom conflict behavior using .on_conflict().
        If no special behavior was specified, this uses the normal Django create(..)

        Arguments:
            fields:
                The fields of the row to create.

            using:
                The name of the database connection
                to use for this query.

        Returns:
            The model instance representing the row that was created.
        """

        if not self.conflict_target and not self.conflict_action:
            # no special action required, use the standard Django create(..)
            return super().create(**fields)

        compiler = self._build_insert_compiler([fields], using=using)
        rows = compiler.execute_sql(return_id=False)

        if not rows:
            return None

        columns = rows[0]

        # get a list of columns that are officially part of the model and
        # preserve the fact that the attribute name
        # might be different than the database column name
        model_columns = {}
        for field in self.model._meta.local_concrete_fields:
            model_columns[field.column] = field.attname

        # strip out any columns/fields returned by the db that
        # are not present in the model
        model_init_fields = {}
        for column_name, column_value in columns.items():
            try:
                model_init_fields[model_columns[column_name]] = column_value
            except KeyError:
                pass

        return self._create_model_instance(model_init_fields, compiler.using)

    def upsert(
        self,
        conflict_target: ConflictTarget,
        fields: dict,
        index_predicate: Optional[str] = None,
        using: Optional[str] = None,
    ) -> int:
        """Creates a new record or updates the existing one with the specified
        data.

        Arguments:
            conflict_target:
                Fields to pass into the ON CONFLICT clause.

            fields:
                Fields to insert/update.

            index_predicate:
                The index predicate to satisfy an arbiter partial index (i.e. what partial index to use for checking
                conflicts)

            using:
                The name of the database connection to
                use for this query.

        Returns:
            The primary key of the row that was created/updated.
        """

        self.on_conflict(
            conflict_target, ConflictAction.UPDATE, index_predicate
        )
        return self.insert(**fields, using=using)

    def upsert_and_get(
        self,
        conflict_target: ConflictTarget,
        fields: dict,
        index_predicate: Optional[str] = None,
        using: Optional[str] = None,
    ):
        """Creates a new record or updates the existing one with the specified
        data and then gets the row.

        Arguments:
            conflict_target:
                Fields to pass into the ON CONFLICT clause.

            fields:
                Fields to insert/update.

            index_predicate:
                The index predicate to satisfy an arbiter partial index (i.e. what partial index to use for checking
                conflicts)

            using:
                The name of the database connection to
                use for this query.

        Returns:
            The model instance representing the row
            that was created/updated.
        """

        self.on_conflict(
            conflict_target, ConflictAction.UPDATE, index_predicate
        )
        return self.insert_and_get(**fields, using=using)

    def bulk_upsert(
        self,
        conflict_target: ConflictTarget,
        rows: Iterable[Dict],
        index_predicate: str = None,
        return_model: bool = False,
        using: Optional[str] = None,
    ):
        """Creates a set of new records or updates the existing ones with the
        specified data.

        Arguments:
            conflict_target:
                Fields to pass into the ON CONFLICT clause.

            rows:
                Rows to upsert.

            index_predicate:
                The index predicate to satisfy an arbiter partial index (i.e. what partial index to use for checking
                conflicts)

            return_model (default: False):
                If model instances should be returned rather than
                just dicts.

            using:
                The name of the database connection to use
                for this query.

        Returns:
            A list of either the dicts of the rows upserted, including the pk or
            the models of the rows upserted
        """

        def is_empty(r):
            return all([False for _ in r])

        if not rows or is_empty(rows):
            return []

        self.on_conflict(
            conflict_target, ConflictAction.UPDATE, index_predicate
        )
        return self.bulk_insert(rows, return_model, using=using)

    def _create_model_instance(self, field_values, using: Optional[str] = None):
        """Creates a new instance of the model with the specified field.

        Use this after the row was inserted into the database. The new
        instance will marked as "saved".
        """

        instance = self.model(**field_values)
        instance._state.db = using
        instance._state.adding = False

        return instance

    def _build_insert_compiler(
        self, rows: Iterable[Dict], using: Optional[str] = None
    ):
        """Builds the SQL compiler for a insert query.

        Arguments:
            rows:
                An iterable of dictionaries, where each entry
                describes a record to insert.

            using:
                The name of the database connection to use
                for this query.

        Returns:
            The SQL compiler for the insert.
        """

        # ask the db router which connection to use
        using = (
            using or self._db or router.db_for_write(self.model, **self._hints)
        )

        # create model objects, we also have to detect cases
        # such as:
        #   [dict(first_name='swen'), dict(fist_name='swen', last_name='kooij')]
        # we need to be certain that each row specifies the exact same
        # amount of fields/columns
        objs = []
        rows_iter = iter(rows)
        first_row = next(rows_iter)
        field_count = len(first_row)
        for index, row in enumerate(chain([first_row], rows_iter)):
            if field_count != len(row):
                raise SuspiciousOperation(
                    (
                        "In bulk upserts, you cannot have rows with different field "
                        "configurations. Row {0} has a different field config than "
                        "the first row."
                    ).format(index)
                )

            objs.append(self._create_model_instance(row, using))

        # get the fields to be used during update/insert
        insert_fields, update_fields = self._get_upsert_fields(first_row)

        # build a normal insert query
        query = PostgresInsertQuery(self.model)
        query.conflict_action = self.conflict_action
        query.conflict_target = self.conflict_target
        query.index_predicate = self.index_predicate
        query.values(objs, insert_fields, update_fields)

        compiler = query.get_compiler(using)
        return compiler

    def _is_magical_field(self, model_instance, field, is_insert: bool):
        """Verifies whether this field is gonna modify something on its own.

        "Magical" means that a field modifies the field value
        during the pre_save.

        Arguments:
            model_instance:
                The model instance the field is defined on.

            field:
                The field to get of whether the field is
                magical.

            is_insert:
                Pretend whether this is an insert?

        Returns:
            True when this field modifies something.
        """

        # does this field modify someting upon insert?
        old_value = getattr(model_instance, field.name, None)
        field.pre_save(model_instance, is_insert)
        new_value = getattr(model_instance, field.name, None)

        return old_value != new_value

    def _get_upsert_fields(self, kwargs):
        """Gets the fields to use in an upsert.

        This some nice magic. We'll split the fields into
        a group of "insert fields" and "update fields":

        INSERT INTO bla ("val1", "val2") ON CONFLICT DO UPDATE SET val1 = EXCLUDED.val1

                         ^^^^^^^^^^^^^^                            ^^^^^^^^^^^^^^^^^^^^
                         insert_fields                                 update_fields

        Often, fields appear in both lists. But, for example,
        a :see:DateTime field with `auto_now_add=True` set, will
        only appear in "insert_fields", since it won't be set
        on existing rows.

        Other than that, the user specificies a list of fields
        in the upsert() call. That migt not be all fields. The
        user could decide to leave out optional fields. If we
        end up doing an update, we don't want to overwrite
        those non-specified fields.

        We cannot just take the list of fields the user
        specifies, because as mentioned, some fields
        make modifications to the model on their own.

        We'll have to detect which fields make modifications
        and include them in the list of insert/update fields.
        """

        model_instance = self.model(**kwargs)
        insert_fields = []
        update_fields = []

        for field in model_instance._meta.local_concrete_fields:
            has_default = field.default != NOT_PROVIDED
            if field.name in kwargs or field.column in kwargs:
                insert_fields.append(field)
                update_fields.append(field)
                continue
            elif has_default:
                insert_fields.append(field)
                continue

            # special handling for 'pk' which always refers to
            # the primary key, so if we the user specifies `pk`
            # instead of a concrete field, we have to handle that
            if field.primary_key is True and "pk" in kwargs:
                insert_fields.append(field)
                update_fields.append(field)
                continue

            if self._is_magical_field(model_instance, field, is_insert=True):
                insert_fields.append(field)

            if self._is_magical_field(model_instance, field, is_insert=False):
                update_fields.append(field)

        return insert_fields, update_fields



from django.db.migrations.operations.models import CreateModel

from psqlextra.backend.migrations.state import (
    PostgresMaterializedViewModelState,
)


class PostgresCreateMaterializedViewModel(CreateModel):
    """Creates the model as a native PostgreSQL 11.x materialzed view."""

    serialization_expand_args = [
        "fields",
        "options",
        "managers",
        "view_options",
    ]

    def __init__(
        self,
        name,
        fields,
        options=None,
        view_options={},
        bases=None,
        managers=None,
    ):
        super().__init__(name, fields, options, bases, managers)

        self.view_options = view_options or {}

    def state_forwards(self, app_label, state):
        state.add_model(
            PostgresMaterializedViewModelState(
                app_label=app_label,
                name=self.name,
                fields=list(self.fields),
                options=dict(self.options),
                bases=tuple(self.bases),
                managers=list(self.managers),
                view_options=dict(self.view_options),
            )
        )

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        """Apply this migration operation forwards."""

        model = to_state.apps.get_model(app_label, self.name)
        if self.allow_migrate_model(schema_editor.connection.alias, model):
            schema_editor.create_materialized_view_model(model)

    def database_backwards(
        self, app_label, schema_editor, from_state, to_state
    ):
        """Apply this migration operation backwards."""

        model = from_state.apps.get_model(app_label, self.name)
        if self.allow_migrate_model(schema_editor.connection.alias, model):
            schema_editor.delete_materialized_view_model(model)

    def deconstruct(self):
        name, args, kwargs = super().deconstruct()

        if self.view_options:
            kwargs["view_options"] = self.view_options

        return name, args, kwargs

    def describe(self):
        """Gets a human readable text describing this migration."""

        description = super().describe()
        description = description.replace("model", "materialized view model")
        return description



from typing import Optional

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import connections
from django.db.models import Manager

from psqlextra.query import PostgresQuerySet


class PostgresManager(Manager.from_queryset(PostgresQuerySet)):
    """Adds support for PostgreSQL specifics."""

    use_in_migrations = True

    def __init__(self, *args, **kwargs):
        """Initializes a new instance of :see:PostgresManager."""

        super().__init__(*args, **kwargs)

        # make sure our back-end is set in at least one db and refuse to proceed
        has_psqlextra_backend = any(
            [
                db_settings
                for db_settings in settings.DATABASES.values()
                if "psqlextra" in db_settings["ENGINE"]
            ]
        )

        if not has_psqlextra_backend:
            raise ImproperlyConfigured(
                (
                    "Could not locate the 'psqlextra.backend'. "
                    "django-postgres-extra cannot function without "
                    "the 'psqlextra.backend'. Set DATABASES.ENGINE."
                )
            )

    def truncate(
        self, cascade: bool = False, using: Optional[str] = None
    ) -> None:
        """Truncates this model/table using the TRUNCATE statement.

        This DELETES ALL ROWS. No signals will be fired.

        See: https://www.postgresql.org/docs/9.1/sql-truncate.html

        Arguments:
            cascade:
                Whether to delete dependent rows. If set to
                False, an error will be raised if there
                are rows in other tables referencing
                the rows you're trying to delete.
        """

        connection = connections[using or "default"]
        table_name = connection.ops.quote_name(self.model._meta.db_table)

        with connection.cursor() as cursor:
            sql = "TRUNCATE TABLE %s" % table_name
            if cascade:
                sql += " CASCADE"

            cursor.execute(sql)



"""Thin wrappers over functions in connection.introspection that don't require
creating a cursor.

This makes test code less verbose and easier to read/write.
"""

from django.db import connection


def table_names(include_views: bool = True):
    """Gets a flat list of tables in the default database."""

    with connection.cursor() as cursor:
        introspection = connection.introspection
        return introspection.table_names(cursor, include_views)


def get_partitioned_table(table_name: str):
    """Gets the definition of a partitioned table in the default database."""

    with connection.cursor() as cursor:
        introspection = connection.introspection
        return introspection.get_partitioned_table(cursor, table_name)


def get_partitions(table_name: str):
    """Gets a list of partitions for the specified partitioned table in the
    default database."""

    with connection.cursor() as cursor:
        introspection = connection.introspection
        return introspection.get_partitions(cursor, table_name)


def get_constraints(table_name: str):
    """Gets a complete list of constraints and indexes for the specified
    table."""

    with connection.cursor() as cursor:
        introspection = connection.introspection
        return introspection.get_constraints(cursor, table_name)



import pytest

from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.test import override_settings

from psqlextra.manager import PostgresManager
from psqlextra.models import PostgresModel

from .fake_model import get_fake_model


@pytest.mark.parametrize(
    "databases",
    [
        {"default": {"ENGINE": "psqlextra.backend"}},
        {
            "default": {"ENGINE": "django.db.backends.postgresql"},
            "other": {"ENGINE": "psqlextra.backend"},
        },
        {
            "default": {"ENGINE": "psqlextra.backend"},
            "other": {"ENGINE": "psqlextra.backend"},
        },
    ],
)
def test_manager_backend_set(databases):
    """Tests that creating a new instance of :see:PostgresManager succeseeds
    without any errors if one or more databases are configured with
    `psqlextra.backend` as its ENGINE."""

    with override_settings(DATABASES=databases):
        assert PostgresManager()


def test_manager_backend_not_set():
    """Tests whether creating a new instance of
    :see:PostgresManager fails if no database
    has `psqlextra.backend` configured
    as its ENGINE."""

    with override_settings(
        DATABASES={"default": {"ENGINE": "django.db.backends.postgresql"}}
    ):
        with pytest.raises(ImproperlyConfigured):
            PostgresManager()


def test_manager_truncate():
    """Tests whether truncating a table works."""

    model = get_fake_model({"name": models.CharField(max_length=255)})

    model.objects.create(name="henk1")
    model.objects.create(name="henk2")

    assert model.objects.count() == 2
    model.objects.truncate()
    assert model.objects.count() == 0


@pytest.mark.django_db(transaction=True)
def test_manager_truncate_cascade():
    """Tests whether truncating a table with cascade works."""

    model_1 = get_fake_model({"name": models.CharField(max_length=255)})

    model_2 = get_fake_model(
        {
            "name": models.CharField(max_length=255),
            "model_1": models.ForeignKey(
                model_1, on_delete=models.CASCADE, null=True
            ),
        }
    )

    obj_1 = model_1.objects.create(name="henk1")
    model_2.objects.create(name="henk1", model_1_id=obj_1.id)

    assert model_1.objects.count() == 1
    assert model_2.objects.count() == 1

    model_1.objects.truncate(cascade=True)

    assert model_1.objects.count() == 0
    assert model_2.objects.count() == 0


def test_manager_truncate_quote_name():
    """Tests whether the truncate statement properly quotes the table name."""

    model = get_fake_model(
        {"name": models.CharField(max_length=255)},
        PostgresModel,
        {
            # without quoting, table names are always
            # lower-case, using a capital case table
            # name requires quoting to work
            "db_table": "MyTable"
        },
    )

    model.objects.create(name="henk1")
    model.objects.create(name="henk2")

    assert model.objects.count() == 2
    model.objects.truncate()
    assert model.objects.count() == 0



from contextlib import contextmanager

from .patched_autodetector import patched_autodetector
from .patched_project_state import patched_project_state


@contextmanager
def postgres_patched_migrations():
    """Patches migration related classes/functions to extend how Django
    generates and applies migrations.

    This adds support for automatically detecting changes in Postgres
    specific models.
    """

    with patched_project_state():
        with patched_autodetector():
            yield



from .add_default_partition import PostgresAddDefaultPartition
from .add_list_partition import PostgresAddListPartition
from .add_range_partition import PostgresAddRangePartition
from .apply_state import ApplyState
from .create_materialized_view_model import PostgresCreateMaterializedViewModel
from .create_partitioned_model import PostgresCreatePartitionedModel
from .create_view_model import PostgresCreateViewModel
from .delete_default_partition import PostgresDeleteDefaultPartition
from .delete_list_partition import PostgresDeleteListPartition
from .delete_materialized_view_model import PostgresDeleteMaterializedViewModel
from .delete_partitioned_model import PostgresDeletePartitionedModel
from .delete_range_partition import PostgresDeleteRangePartition
from .delete_view_model import PostgresDeleteViewModel

__all__ = [
    "ApplyState",
    "PostgresAddRangePartition",
    "PostgresAddListPartition",
    "PostgresAddDefaultPartition",
    "PostgresDeleteDefaultPartition",
    "PostgresDeleteRangePartition",
    "PostgresDeleteListPartition",
    "PostgresCreatePartitionedModel",
    "PostgresDeletePartitionedModel",
    "PostgresCreateViewModel",
    "PostgresCreateMaterializedViewModel",
    "PostgresDeleteViewModel",
    "PostgresDeleteMaterializedViewModel",
]



import django
import pytest

from django.db import models

from psqlextra.fields import HStoreField

from .fake_model import get_fake_model


@pytest.fixture
def model():
    """Test models, where the first model has a foreign key relationship to the
    second."""

    return get_fake_model({"title": HStoreField()})


@pytest.fixture
def modelobj(model):
    """Data for the test models, one row per model."""

    return model.objects.create(title={"en": "english", "ar": "arabic"})


def test_query_values_hstore(model, modelobj):
    """Tests that selecting all the keys properly works and returns a.

    :see:LocalizedValue instance.
    """

    result = list(model.objects.values("title"))[0]
    assert result["title"] == modelobj.title


def test_query_values_hstore_key(model, modelobj):
    """Tests whether selecting a single key from a :see:HStoreField using the
    query set's .values() works properly."""

    result = list(model.objects.values("title__en", "title__ar"))[0]
    assert result["title__en"] == modelobj.title["en"]
    assert result["title__ar"] == modelobj.title["ar"]


def test_query_values_list_hstore_key(model, modelobj):
    """Tests that selecting a single key from a :see:HStoreField using the
    query set's .values_list() works properly."""

    result = list(model.objects.values_list("title__en", "title__ar"))[0]
    assert result[0] == modelobj.title["en"]
    assert result[1] == modelobj.title["ar"]


def test_query_values_hstore_key_through_fk():
    """Tests whether selecting a single key from a :see:HStoreField using the
    query set's .values() works properly when there's a foreign key
    relationship involved."""

    # this starting working in django 2.1
    # see: https://github.com/django/django/commit/20bab2cf9d02a5c6477d8aac066a635986e0d3f3
    if django.VERSION < (2, 1):
        return

    fmodel = get_fake_model({"name": HStoreField()})

    model = get_fake_model(
        {"fk": models.ForeignKey(fmodel, on_delete=models.CASCADE)}
    )

    fobj = fmodel.objects.create(name={"en": "swen", "ar": "arabic swen"})
    model.objects.create(fk=fobj)

    result = list(model.objects.values("fk__name__ar"))[0]
    assert result["fk__name__ar"] == fobj.name["ar"]



from .base import PostgresModel
from .partitioned import PostgresPartitionedModel
from .view import PostgresMaterializedViewModel, PostgresViewModel

__all__ = [
    "PostgresModel",
    "PostgresViewModel",
    "PostgresMaterializedViewModel",
    "PostgresPartitionedModel",
]



from psqlextra.backend.migrations.state import PostgresRangePartitionState

from .partition import PostgresPartitionOperation


class PostgresAddRangePartition(PostgresPartitionOperation):
    """Adds a new range partition to a :see:PartitionedPostgresModel."""

    def __init__(self, model_name: str, name: str, from_values, to_values):
        """Initializes new instance of :see:AddRangePartition.

        Arguments:
            model_name:
                The name of the :see:PartitionedPostgresModel.

            name:
                The name to give to the new partition table.

            from_values:
                Start of the partitioning key range of
                values that need to be stored in this
                partition.

            to_values:
                End of the partitioning key range of
                values that need to be stored in this
                partition.
        """

        super().__init__(model_name, name)

        self.from_values = from_values
        self.to_values = to_values

    def state_forwards(self, app_label, state):
        model = state.models[(app_label, self.model_name_lower)]
        model.add_partition(
            PostgresRangePartitionState(
                app_label=app_label,
                model_name=self.model_name,
                name=self.name,
                from_values=self.from_values,
                to_values=self.to_values,
            )
        )

        state.reload_model(app_label, self.model_name_lower)

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        model = to_state.apps.get_model(app_label, self.model_name)
        if self.allow_migrate_model(schema_editor.connection.alias, model):
            schema_editor.add_range_partition(
                model, self.name, self.from_values, self.to_values
            )

    def deconstruct(self):
        name, args, kwargs = super().deconstruct()

        kwargs["from_values"] = self.from_values
        kwargs["to_values"] = self.to_values

        return name, args, kwargs

    def describe(self) -> str:
        return "Creates range partition %s on %s" % (self.name, self.model_name)



from .partition import PostgresPartitionOperation


class PostgresDeletePartition(PostgresPartitionOperation):
    """Deletes a partition that's part of a :see:PartitionedPostgresModel."""

    def state_forwards(self, app_label, state):
        model = state.models[(app_label, self.model_name_lower)]
        model.delete_partition(self.name)

        state.reload_model(app_label, self.model_name_lower)

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        model = from_state.apps.get_model(app_label, self.model_name)
        if self.allow_migrate_model(schema_editor.connection.alias, model):
            schema_editor.delete_partition(model, self.name)

    def database_backwards(
        self, app_label, schema_editor, from_state, to_state
    ):
        model = to_state.apps.get_model(app_label, self.model_name)
        model_state = to_state.models[(app_label, self.model_name)]

        if self.allow_migrate_model(schema_editor.connection.alias, model):
            partition_state = model_state.partitions[self.name]
            schema_editor.add_default_partition(model, partition_state.name)

    def describe(self) -> str:
        return "Deletes partition %s on %s" % (self.name, self.model_name)



from .hstore_required import HStoreRequiredSchemaEditorSideEffect
from .hstore_unique import HStoreUniqueSchemaEditorSideEffect

__all__ = [
    "HStoreUniqueSchemaEditorSideEffect",
    "HStoreRequiredSchemaEditorSideEffect",
]



from psqlextra.models import PostgresPartitionedModel
from psqlextra.types import PostgresPartitioningMethod

from .fake_model import define_fake_partitioned_model


def test_partitioned_model_abstract():
    """Tests whether :see:PostgresPartitionedModel is abstract."""

    assert PostgresPartitionedModel._meta.abstract


def test_partitioning_model_options_meta():
    """Tests whether the `_partitioning_meta` attribute is available on the
    class (created by the meta class) and not just creating when the model is
    instantiated."""

    assert PostgresPartitionedModel._partitioning_meta


def test_partitioned_model_default_options():
    """Tests whether the default partitioning options are set as expected on.

    :see:PostgresPartitionedModel.
    """

    model = define_fake_partitioned_model()

    assert model._partitioning_meta.method == PostgresPartitioningMethod.RANGE
    assert model._partitioning_meta.key == []


def test_partitioned_model_method_option():
    """Tests whether the `method` partitioning option is properly copied onto
    the options object."""

    model = define_fake_partitioned_model(
        partitioning_options=dict(method=PostgresPartitioningMethod.LIST)
    )

    assert model._partitioning_meta.method == PostgresPartitioningMethod.LIST


def test_partitioned_model_method_option_none():
    """Tests whether setting the `method` partitioning option results in the
    default being set."""

    model = define_fake_partitioned_model(
        partitioning_options=dict(method=None)
    )

    assert model._partitioning_meta.method == PostgresPartitioningMethod.RANGE


def test_partitioned_model_key_option():
    """Tests whether the `key` partitioning option is properly copied onto the
    options object."""

    model = define_fake_partitioned_model(
        partitioning_options=dict(key=["timestamp"])
    )

    assert model._partitioning_meta.key == ["timestamp"]


def test_partitioned_model_key_option_none():
    """Tests whether setting the `key` partitioning option results in the
    default being set."""

    model = define_fake_partitioned_model(partitioning_options=dict(key=None))

    assert model._partitioning_meta.key == []



from abc import abstractmethod
from typing import Generator

from .partition import PostgresPartition


class PostgresPartitioningStrategy:
    """Base class for implementing a partitioning strategy for a partitioned
    table."""

    @abstractmethod
    def to_create(self,) -> Generator[PostgresPartition, None, None]:
        """Generates a list of partitions to be created."""

    @abstractmethod
    def to_delete(self,) -> Generator[PostgresPartition, None, None]:
        """Generates a list of partitions to be deleted."""


__all__ = ["PostgresPartitioningStrategy"]



import sys

from typing import Optional

import colorama

from ansimarkup import ansiprint, ansistring
from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils.module_loading import import_string

from psqlextra.partitioning import PostgresPartitioningError


class Command(BaseCommand):
    """Create new partitions and delete old ones according to the configured
    partitioning strategies."""

    help = "Create new partitions and delete old ones using the configured partitioning manager. The PSQLEXTRA_PARTITIONING_MANAGER setting must be configured."

    def add_arguments(self, parser):
        parser.add_argument(
            "--dry",
            "-d",
            action="store_true",
            help="When specified, no partition will be created/deleted. Just a simulation.",
            required=False,
            default=False,
        )

        parser.add_argument(
            "--yes",
            "-y",
            action="store_true",
            help="Answer yes to all questions. WARNING: You will not be asked before deleting a partition.",
            required=False,
            default=False,
        )

        parser.add_argument(
            "--using",
            "-u",
            help="Name of the database connection to use.",
            default="default",
        )

        parser.add_argument(
            "--skip-create",
            action="store_true",
            help="Do not create partitions.",
            required=False,
            default=False,
        )

        parser.add_argument(
            "--skip-delete",
            action="store_true",
            help="Do not delete partitions.",
            required=False,
            default=False,
        )

    def handle(
        self,
        dry: bool,
        yes: bool,
        using: Optional[str],
        skip_create: bool,
        skip_delete: bool,
        *args,
        **kwargs,
    ):
        # disable coloring if no terminal is attached
        if not sys.stdout.isatty():
            colorama.init(strip=True)

        partitioning_manager = self._partitioning_manager()

        plan = partitioning_manager.plan(
            skip_create=skip_create, skip_delete=skip_delete, using=using
        )

        creations_count = len(plan.creations)
        deletions_count = len(plan.deletions)
        if creations_count == 0 and deletions_count == 0:
            ansiprint("<b><white>Nothing to be done.</white></b>")
            return

        plan.print()

        if dry:
            return

        if not yes:
            sys.stdout.write(
                ansistring(
                    "<b><white>Do you want to proceed? (y/N) </white></b>"
                )
            )

            if not self._ask_for_confirmation():
                ansiprint("<b><white>Operation aborted.</white></b>")
                return

        plan.apply(using=using)
        ansiprint("<b><white>Operations applied.</white></b>")

    @staticmethod
    def _ask_for_confirmation() -> bool:
        answer = input("").lower()
        if not answer:
            return False

        if answer[0] == "y" or answer == "yes":
            return True

        return False

    @staticmethod
    def _partitioning_manager():
        partitioning_manager = getattr(
            settings, "PSQLEXTRA_PARTITIONING_MANAGER"
        )
        if not partitioning_manager:
            raise PostgresPartitioningError(
                "You must configure the PSQLEXTRA_PARTITIONING_MANAGER setting "
                "for automatic partitioning to work."
            )

        if isinstance(partitioning_manager, str):
            partitioning_manager = import_string(partitioning_manager)

        return partitioning_manager



from django.db import models
from django.db.models import F

from psqlextra.expressions import HStoreRef
from psqlextra.fields import HStoreField

from .fake_model import get_fake_model


def test_query_annotate_hstore_key_ref():
    """Tests whether annotating using a :see:HStoreRef expression works
    correctly.

    This allows you to select an individual hstore key.
    """

    model_fk = get_fake_model({"title": HStoreField()})

    model = get_fake_model(
        {"fk": models.ForeignKey(model_fk, on_delete=models.CASCADE)}
    )

    fk = model_fk.objects.create(title={"en": "english", "ar": "arabic"})
    model.objects.create(fk=fk)

    queryset = (
        model.objects.annotate(english_title=HStoreRef("fk__title", "en"))
        .values("english_title")
        .first()
    )

    assert queryset["english_title"] == "english"


def test_query_annotate_rename():
    """Tests whether field names can be overwritten with a annotated field."""

    model = get_fake_model({"title": models.CharField(max_length=12)})

    model.objects.create(title="swen")

    obj = model.objects.annotate(title=F("title")).first()
    assert obj.title == "swen"


def test_query_annotate_rename_chain():
    """Tests whether annotations are behaving correctly after a QuerySet
    chain."""

    model = get_fake_model(
        {
            "name": models.CharField(max_length=10),
            "value": models.IntegerField(),
        }
    )

    model.objects.create(name="test", value=23)

    obj = model.objects.values("name").annotate(value=F("value"))[:1]
    assert "value" in obj[0]
    assert obj[0]["value"] == 23


def test_query_annotate_rename_order():
    """Tests whether annotation order is preserved after a rename."""

    model = get_fake_model(
        {
            "name": models.CharField(max_length=10),
            "value": models.IntegerField(),
        }
    )

    qs = model.objects.annotate(value=F("value"), value_2=F("value"))
    assert list(qs.query.annotations.keys()) == ["value", "value_2"]


def test_query_hstore_value_update_f_ref():
    """Tests whether F(..) expressions can be used in hstore values when
    performing update queries."""

    model = get_fake_model(
        {"name": models.CharField(max_length=255), "name_new": HStoreField()}
    )

    model.objects.create(name="waqas", name_new=dict(en="swen"))
    model.objects.update(name_new=dict(en=models.F("name")))

    inst = model.objects.all().first()
    assert inst.name_new.get("en") == "waqas"


def test_query_hstore_value_update_cast():
    """Tests whether values in a HStore field are automatically cast to strings
    when doing updates."""

    model = get_fake_model({"title": HStoreField()})

    model.objects.create(title=dict(en="test"))
    model.objects.update(title=dict(en=2))

    inst = model.objects.all().first()
    assert inst.title.get("en") == "2"


def test_query_hstore_value_update_escape():
    """Tests whether values in a HStore field are properly escaped using
    prepared statement values."""

    model = get_fake_model({"title": HStoreField()})

    model.objects.create(title=dict(en="test"))
    model.objects.update(title=dict(en="console.log('test')"))

    inst = model.objects.all().first()
    assert inst.title.get("en") == "console.log('test')"



import pytest

from django.db import models

from psqlextra.fields import HStoreField
from psqlextra.query import ConflictAction

from .fake_model import get_fake_model


def test_on_conflict_update():
    """Tests whether simple upserts works correctly."""

    model = get_fake_model(
        {
            "title": HStoreField(uniqueness=["key1"]),
            "cookies": models.CharField(max_length=255, null=True),
        }
    )

    obj1 = model.objects.on_conflict(
        [("title", "key1")], ConflictAction.UPDATE
    ).insert_and_get(title={"key1": "beer"}, cookies="cheers")

    obj1.refresh_from_db()
    assert obj1.title["key1"] == "beer"
    assert obj1.cookies == "cheers"

    obj2 = model.objects.on_conflict(
        [("title", "key1")], ConflictAction.UPDATE
    ).insert_and_get(title={"key1": "beer"}, cookies="choco")

    obj1.refresh_from_db()
    obj2.refresh_from_db()

    # assert both objects are the same
    assert obj1.id == obj2.id
    assert obj1.title["key1"] == "beer"
    assert obj1.cookies == "choco"
    assert obj2.title["key1"] == "beer"
    assert obj2.cookies == "choco"


def test_on_conflict_update_foreign_key_by_object():
    """Tests whether simple upsert works correctly when the conflicting field
    is a foreign key specified as an object."""

    other_model = get_fake_model({})

    model = get_fake_model(
        {
            "other": models.OneToOneField(
                other_model, on_delete=models.CASCADE
            ),
            "data": models.CharField(max_length=255),
        }
    )

    other_obj = other_model.objects.create()

    obj1 = model.objects.on_conflict(
        ["other"], ConflictAction.UPDATE
    ).insert_and_get(other=other_obj, data="some data")

    assert obj1.other == other_obj
    assert obj1.data == "some data"

    obj1.refresh_from_db()
    assert obj1.other == other_obj
    assert obj1.data == "some data"

    with pytest.raises(ValueError):
        (
            model.objects.on_conflict(
                ["other"], ConflictAction.UPDATE
            ).insert_and_get(other=obj1)
        )

    obj2 = model.objects.on_conflict(
        ["other"], ConflictAction.UPDATE
    ).insert_and_get(other=other_obj, data="different data")

    assert obj2.other == other_obj
    assert obj2.data == "different data"

    obj1.refresh_from_db()
    obj2.refresh_from_db()

    # assert that the 'other' field didn't change
    assert obj1.id == obj2.id
    assert obj1.other == other_obj
    assert obj2.other == other_obj
    assert obj1.data == "different data"
    assert obj2.data == "different data"


def test_on_conflict_update_foreign_key_by_id():
    """Tests whether simple upsert works correctly when the conflicting field
    is a foreign key specified as an id."""

    other_model = get_fake_model({})

    model = get_fake_model(
        {
            "other": models.OneToOneField(
                other_model, on_delete=models.CASCADE
            ),
            "data": models.CharField(max_length=255),
        }
    )

    other_obj = other_model.objects.create()

    obj1 = model.objects.on_conflict(
        ["other_id"], ConflictAction.UPDATE
    ).insert_and_get(other_id=other_obj.pk, data="some data")

    assert obj1.other == other_obj
    assert obj1.data == "some data"

    obj1.refresh_from_db()
    assert obj1.other == other_obj
    assert obj1.data == "some data"

    obj2 = model.objects.on_conflict(
        ["other_id"], ConflictAction.UPDATE
    ).insert_and_get(other_id=other_obj.pk, data="different data")

    assert obj2.other == other_obj
    assert obj2.data == "different data"

    obj1.refresh_from_db()
    obj2.refresh_from_db()

    # assert that the 'other' field didn't change
    assert obj1.id == obj2.id
    assert obj1.other == other_obj
    assert obj2.other == other_obj
    assert obj1.data == "different data"
    assert obj2.data == "different data"



from django.db.models.indexes import Index


class CaseInsensitiveUniqueIndex(Index):
    sql_create_unique_index = (
        "CREATE UNIQUE INDEX %(name)s ON %(table)s (%(columns)s)%(extra)s"
    )

    def create_sql(self, model, schema_editor, using="", **kwargs):
        statement = super().create_sql(model, schema_editor, using)
        statement.template = self.sql_create_unique_index

        column_collection = statement.parts["columns"]
        statement.parts["columns"] = ", ".join(
            [
                "LOWER(%s)" % self._quote_column(column_collection, column, idx)
                for idx, column in enumerate(column_collection.columns)
            ]
        )

        return statement

    def deconstruct(self):
        """Serializes the :see:CaseInsensitiveUniqueIndex for the migrations
        file."""
        _, args, kwargs = super().deconstruct()
        path = "%s.%s" % (self.__class__.__module__, self.__class__.__name__)
        path = path.replace("django.db.models.indexes", "django.db.models")

        return path, args, kwargs

    @staticmethod
    def _quote_column(column_collection, column, idx):
        quoted_name = column_collection.quote_name(column)
        try:
            return quoted_name + column_collection.col_suffixes[idx]
        except IndexError:
            return column_collection.quote_name(column)



from django.db import migrations
from django.db.migrations.autodetector import MigrationAutodetector
from django.db.migrations.state import ProjectState

from psqlextra.fields import HStoreField


def _make_project_state(model_states):
    """Shortcut to make :see:ProjectState from a list of predefined models."""

    project_state = ProjectState()
    for model_state in model_states:
        project_state.add_model(model_state.clone())
    return project_state


def _detect_changes(before_states, after_states):
    """Uses the migration autodetector to detect changes in the specified
    project states."""

    return MigrationAutodetector(
        _make_project_state(before_states), _make_project_state(after_states)
    )._detect_changes()


def _assert_autodetector(changes, expected):
    """Asserts whether the results of the auto detector are as expected."""

    assert "tests" in changes
    assert len("tests") > 0

    operations = changes["tests"][0].operations

    for i, expected_operation in enumerate(expected):
        real_operation = operations[i]
        _, _, real_args, real_kwargs = real_operation.field.deconstruct()
        _, _, expected_args, expected_kwargs = (
            expected_operation.field.deconstruct()
        )

        assert real_args == expected_args
        assert real_kwargs == expected_kwargs


def test_hstore_autodetect_uniqueness():
    """Tests whether changes in the `uniqueness` option are properly detected
    by the auto detector."""

    before = [
        migrations.state.ModelState(
            "tests", "Model1", [("title", HStoreField())]
        )
    ]
    after = [
        migrations.state.ModelState(
            "tests", "Model1", [("title", HStoreField(uniqueness=["en"]))]
        )
    ]

    changes = _detect_changes(before, after)

    _assert_autodetector(
        changes,
        [
            migrations.AlterField(
                "Model1", "title", HStoreField(uniqueness=["en"])
            )
        ],
    )


def test_hstore_autodetect_required():
    """Tests whether changes in the `required` option are properly detected by
    the auto detector."""

    before = [
        migrations.state.ModelState(
            "tests", "Model1", [("title", HStoreField())]
        )
    ]
    after = [
        migrations.state.ModelState(
            "tests", "Model1", [("title", HStoreField(required=["en"]))]
        )
    ]

    changes = _detect_changes(before, after)

    _assert_autodetector(
        changes,
        [
            migrations.AlterField(
                "Model1", "title", HStoreField(required=["en"])
            )
        ],
    )



from typing import Type

from psqlextra.models import PostgresViewModel

from .model import PostgresModelState


class PostgresViewModelState(PostgresModelState):
    """Represents the state of a :see:PostgresViewModel in the migrations."""

    def __init__(self, *args, view_options={}, **kwargs):
        """Initializes a new instance of :see:PostgresViewModelState.

        Arguments:
            view_options:
                Dictionary of options for views.
                See: PostgresViewModelMeta for a list.
        """

        super().__init__(*args, **kwargs)

        self.view_options = dict(view_options)

    @classmethod
    def _pre_new(
        cls, model: PostgresViewModel, model_state: "PostgresViewModelState"
    ) -> "PostgresViewModelState":
        """Called when a new model state is created from the specified
        model."""

        model_state.view_options = dict(model._view_meta.original_attrs)
        return model_state

    def _pre_clone(
        self, model_state: "PostgresViewModelState"
    ) -> "PostgresViewModelState":
        """Called when this model state is cloned."""

        model_state.view_options = dict(self.view_options)
        return model_state

    def _pre_render(self, name: str, bases, attributes):
        """Called when this model state is rendered into a model."""

        view_meta = type("ViewMeta", (), dict(self.view_options))
        return name, bases, {**attributes, "ViewMeta": view_meta}

    @classmethod
    def _get_base_model_class(self) -> Type[PostgresViewModel]:
        """Gets the class to use as a base class for rendered models."""

        return PostgresViewModel



from enum import Enum
from typing import Any, Dict, List, Tuple, Union

SQL = str
SQLWithParams = Tuple[str, Union[Tuple[Any, ...], Dict[str, Any]]]


class StrEnum(str, Enum):
    @classmethod
    def all(cls) -> List["PostgresPartitioningMethod"]:
        return [choice for choice in cls]

    @classmethod
    def values(cls) -> List[str]:
        return [choice.value for choice in cls]

    def __str__(self) -> str:
        return str(self.value)


class ConflictAction(Enum):
    """Possible actions to take on a conflict."""

    NOTHING = "NOTHING"
    UPDATE = "UPDATE"

    @classmethod
    def all(cls) -> List["ConflictAction"]:
        return [choice for choice in cls]


class PostgresPartitioningMethod(StrEnum):
    """Methods of partitioning supported by PostgreSQL 11.x native support for
    table partitioning."""

    RANGE = "range"
    LIST = "list"



from django.db import connection


def test_db_backend_hstore_extension_enabled():
    """Tests whether the `hstore` extension was enabled automatically."""

    with connection.cursor() as cursor:
        cursor.execute(
            ("SELECT count(*) FROM pg_extension " "WHERE extname = 'hstore'")
        )

        assert cursor.fetchone()[0] == 1



import uuid

import pytest

from django.db import models

from psqlextra.query import ConflictAction

from ..fake_model import get_fake_model

ROW_COUNT = 10000


@pytest.mark.benchmark()
def test_upsert_bulk_naive(benchmark):
    model = get_fake_model(
        {"field": models.CharField(max_length=255, unique=True)}
    )

    rows = []
    random_values = []
    for i in range(0, ROW_COUNT):
        random_value = str(uuid.uuid4())
        random_values.append(random_value)
        rows.append(model(field=random_value))

    model.objects.bulk_create(rows)

    def _native_upsert(model, random_values):
        """Performs a concurrency safe upsert using the native PostgreSQL
        upsert."""

        rows = [dict(field=random_value) for random_value in random_values]

        for row in rows:
            model.objects.on_conflict(["field"], ConflictAction.UPDATE).insert(
                **row
            )

    benchmark(_native_upsert, model, random_values)


@pytest.mark.benchmark()
def test_upsert_bulk_native(benchmark):
    model = get_fake_model(
        {"field": models.CharField(max_length=255, unique=True)}
    )

    rows = []
    random_values = []
    for i in range(0, ROW_COUNT):
        random_value = str(uuid.uuid4())
        random_values.append(random_value)
        rows.append(model(field=random_value))

    model.objects.bulk_create(rows)

    def _native_upsert(model, random_values):
        """Performs a concurrency safe upsert using the native PostgreSQL
        upsert."""

        rows = [dict(field=random_value) for random_value in random_values]

        model.objects.on_conflict(["field"], ConflictAction.UPDATE).bulk_insert(
            rows
        )

    benchmark(_native_upsert, model, random_values)



import pytest

from django.db import transaction
from django.db.utils import IntegrityError

from psqlextra.fields import HStoreField

from . import migrations
from .fake_model import get_fake_model


def test_hstore_unique_migration_create_drop_model():
    """Tests whether indexes are properly created and dropped when creating and
    dropping a model."""

    uniqueness = ["beer", "cookies"]

    test = migrations.create_drop_model(
        HStoreField(uniqueness=uniqueness), ["CREATE UNIQUE", "DROP INDEX"]
    )

    with test as calls:
        assert len(calls["CREATE UNIQUE"]) == len(uniqueness)
        assert len(calls["DROP INDEX"]) == len(uniqueness)


def test_hstore_unique_migration_alter_db_table():
    """Tests whether indexes are renamed properly when renaming the database
    table."""

    test = migrations.alter_db_table(
        HStoreField(uniqueness=["beer", "cookie"]),
        ["RENAME TO", "CREATE INDEX", "DROP INDEX"],
    )

    with test as calls:
        # 1 rename for table, 2 for hstore keys
        assert len(calls["RENAME TO"]) == 3
        assert len(calls.get("CREATE UNIQUE", [])) == 0
        assert len(calls.get("DROP INDEX", [])) == 0


def test_hstore_unique_add_field():
    """Tests whether adding a field properly creates the indexes."""

    test = migrations.add_field(
        HStoreField(uniqueness=["beer"]), ["CREATE UNIQUE", "DROP INDEX"]
    )

    with test as calls:
        assert len(calls.get("CREATE UNIQUE", [])) == 1
        assert len(calls.get("DROP INDEX", [])) == 0


def test_hstore_unique_remove_field():
    """Tests whether removing a field properly removes the index."""

    test = migrations.remove_field(
        HStoreField(uniqueness=["beer"]), ["CREATE UNIQUE", "DROP INDEX"]
    )

    with test as calls:
        assert len(calls.get("CREATE UNIQUE", [])) == 0
        assert len(calls.get("DROP INDEX", [])) == 1


def test_hstore_unique_alter_field_nothing():
    """Tests whether no indexes are dropped when not changing anything in the
    uniqueness."""

    test = migrations.alter_field(
        HStoreField(uniqueness=["beer"]),
        HStoreField(uniqueness=["beer"]),
        ["CREATE UNIQUE", "DROP INDEX"],
    )

    with test as calls:
        assert len(calls.get("CREATE UNIQUE", [])) == 0
        assert len(calls.get("DROP INDEX", [])) == 0


def test_hstore_unique_alter_field_add():
    """Tests whether only one index is created when adding another key to the
    uniqueness."""

    test = migrations.alter_field(
        HStoreField(uniqueness=["beer"]),
        HStoreField(uniqueness=["beer", "beer1"]),
        ["CREATE UNIQUE", "DROP INDEX"],
    )

    with test as calls:
        assert len(calls.get("CREATE UNIQUE", [])) == 1
        assert len(calls.get("DROP INDEX", [])) == 0


def test_hstore_unique_alter_field_remove():
    """Tests whether one index is dropped when removing a key from
    uniqueness."""

    test = migrations.alter_field(
        HStoreField(uniqueness=["beer"]),
        HStoreField(uniqueness=[]),
        ["CREATE UNIQUE", "DROP INDEX"],
    )

    with test as calls:
        assert len(calls.get("CREATE UNIQUE", [])) == 0
        assert len(calls.get("DROP INDEX", [])) == 1


def test_hstore_unique_alter_field_add_together():
    """Tests whether adding one index is created when adding a "unique
    together"."""

    test = migrations.alter_field(
        HStoreField(uniqueness=["beer"]),
        HStoreField(uniqueness=["beer", ("beer1", "beer2")]),
        ["CREATE UNIQUE", "DROP INDEX"],
    )

    with test as calls:
        assert len(calls.get("CREATE UNIQUE", [])) == 1
        assert len(calls.get("DROP INDEX", [])) == 0


def test_hstore_unique_alter_field_remove_together():
    """Tests whether adding one index is dropped when adding a "unique
    together"."""

    test = migrations.alter_field(
        HStoreField(uniqueness=[("beer1", "beer2")]),
        HStoreField(uniqueness=[]),
        ["CREATE UNIQUE", "DROP INDEX"],
    )

    with test as calls:
        assert len(calls.get("CREATE UNIQUE", [])) == 0
        assert len(calls.get("DROP INDEX", [])) == 1


def test_hstore_unique_rename_field():
    """Tests whether renaming a field doesn't cause the index to be re-
    created."""

    test = migrations.rename_field(
        HStoreField(uniqueness=["beer", "cookies"]),
        ["RENAME TO", "CREATE INDEX", "DROP INDEX"],
    )

    with test as calls:
        assert len(calls.get("RENAME TO", [])) == 2
        assert len(calls.get("CREATE UNIQUE", [])) == 0
        assert len(calls.get("DROP INDEX", [])) == 0


def test_hstore_unique_enforcement():
    """Tests whether the constraints are actually properly enforced."""
    model = get_fake_model({"title": HStoreField(uniqueness=["en"])})

    # should pass, table is empty and 'ar' does not have to be unique
    model.objects.create(title={"en": "unique", "ar": "notunique"})
    model.objects.create(title={"en": "elseunique", "ar": "notunique"})

    # this should fail, key 'en' must be unique
    with transaction.atomic():
        with pytest.raises(IntegrityError):
            model.objects.create(title={"en": "unique", "ar": "notunique"})


def test_hstore_unique_enforcement_together():
    """Tests whether unique_together style constraints are enforced
    properly."""

    model = get_fake_model({"title": HStoreField(uniqueness=[("en", "ar")])})

    model.objects.create(title={"en": "unique", "ar": "notunique"})

    with transaction.atomic():
        with pytest.raises(IntegrityError):
            model.objects.create(title={"en": "unique", "ar": "notunique"})

    model.objects.create(title={"en": "notunique", "ar": "unique"})



from typing import Any, Optional

from psqlextra.backend.schema import PostgresSchemaEditor
from psqlextra.models import PostgresPartitionedModel

from .partition import PostgresPartition


class PostgresRangePartition(PostgresPartition):
    """Base class for a PostgreSQL table partition in a range partitioned
    table."""

    def __init__(self, from_values: Any, to_values: Any) -> None:
        self.from_values = from_values
        self.to_values = to_values

    def deconstruct(self) -> dict:
        return {
            **super().deconstruct(),
            "from_values": self.from_values,
            "to_values": self.to_values,
        }

    def create(
        self,
        model: PostgresPartitionedModel,
        schema_editor: PostgresSchemaEditor,
        comment: Optional[str] = None,
    ) -> None:
        schema_editor.add_range_partition(
            model=model,
            name=self.name(),
            from_values=self.from_values,
            to_values=self.to_values,
            comment=comment,
        )

    def delete(
        self,
        model: PostgresPartitionedModel,
        schema_editor: PostgresSchemaEditor,
    ) -> None:
        schema_editor.delete_partition(model, self.name())


__all__ = ["PostgresRangePartition"]



from django.db.migrations.operations.base import Operation


class PostgresPartitionOperation(Operation):
    def __init__(self, model_name: str, name: str) -> None:
        """Initializes new instance of :see:AddDefaultPartition.

        Arguments:
            model_name:
                The name of the :see:PartitionedPostgresModel.

            name:
                The name to give to the new partition table.
        """

        self.model_name = model_name
        self.model_name_lower = model_name.lower()
        self.name = name

    def deconstruct(self):
        kwargs = {"model_name": self.model_name, "name": self.name}
        return (self.__class__.__qualname__, [], kwargs)

    def state_forwards(self, *args, **kwargs):
        pass

    def state_backwards(self, *args, **kwargs):
        pass



from django.db import models

from psqlextra.fields import HStoreField

from .fake_model import get_fake_model


def test_upsert():
    """Tests whether simple upserts works correctly."""

    model = get_fake_model(
        {
            "title": HStoreField(uniqueness=["key1"]),
            "cookies": models.CharField(max_length=255, null=True),
        }
    )

    obj1 = model.objects.upsert_and_get(
        conflict_target=[("title", "key1")],
        fields=dict(title={"key1": "beer"}, cookies="cheers"),
    )

    obj1.refresh_from_db()
    assert obj1.title["key1"] == "beer"
    assert obj1.cookies == "cheers"

    obj2 = model.objects.upsert_and_get(
        conflict_target=[("title", "key1")],
        fields=dict(title={"key1": "beer"}, cookies="choco"),
    )

    obj1.refresh_from_db()
    obj2.refresh_from_db()

    # assert both objects are the same
    assert obj1.id == obj2.id
    assert obj1.title["key1"] == "beer"
    assert obj1.cookies == "choco"
    assert obj2.title["key1"] == "beer"
    assert obj2.cookies == "choco"


def test_upsert_explicit_pk():
    """Tests whether upserts works when the primary key is explicitly
    specified."""

    model = get_fake_model(
        {
            "name": models.CharField(max_length=255, primary_key=True),
            "cookies": models.CharField(max_length=255, null=True),
        }
    )

    obj1 = model.objects.upsert_and_get(
        conflict_target=[("name")],
        fields=dict(name="the-object", cookies="first-cheers"),
    )

    obj1.refresh_from_db()
    assert obj1.name == "the-object"
    assert obj1.cookies == "first-cheers"

    obj2 = model.objects.upsert_and_get(
        conflict_target=[("name")],
        fields=dict(name="the-object", cookies="second-boo"),
    )

    obj1.refresh_from_db()
    obj2.refresh_from_db()

    # assert both objects are the same
    assert obj1.pk == obj2.pk
    assert obj1.name == "the-object"
    assert obj1.cookies == "second-boo"
    assert obj2.name == "the-object"
    assert obj2.cookies == "second-boo"


def test_upsert_bulk():
    """Tests whether bulk_upsert works properly."""

    model = get_fake_model(
        {
            "first_name": models.CharField(
                max_length=255, null=True, unique=True
            ),
            "last_name": models.CharField(max_length=255, null=True),
        }
    )

    model.objects.bulk_upsert(
        conflict_target=["first_name"],
        rows=[
            dict(first_name="Swen", last_name="Kooij"),
            dict(first_name="Henk", last_name="Test"),
        ],
    )

    row_a = model.objects.get(first_name="Swen")
    row_b = model.objects.get(first_name="Henk")

    model.objects.bulk_upsert(
        conflict_target=["first_name"],
        rows=[
            dict(first_name="Swen", last_name="Test"),
            dict(first_name="Henk", last_name="Kooij"),
        ],
    )

    row_a.refresh_from_db()
    assert row_a.last_name == "Test"

    row_b.refresh_from_db()
    assert row_b.last_name == "Kooij"


def test_upsert_bulk_no_rows():
    """Tests whether bulk_upsert doesn't crash when specifying no rows or a
    falsy value."""

    model = get_fake_model(
        {"name": models.CharField(max_length=255, null=True, unique=True)}
    )

    model.objects.bulk_upsert(conflict_target=["name"], rows=[])

    model.objects.bulk_upsert(conflict_target=["name"], rows=None)


def test_bulk_upsert_return_models():
    """Tests whether models are returned instead of dictionaries when
    specifying the return_model=True argument."""

    model = get_fake_model(
        {
            "id": models.BigAutoField(primary_key=True),
            "name": models.CharField(max_length=255, unique=True),
        }
    )

    rows = [dict(name="John Smith"), dict(name="Jane Doe")]

    objs = model.objects.bulk_upsert(
        conflict_target=["name"], rows=rows, return_model=True
    )

    for index, obj in enumerate(objs, 1):
        assert isinstance(obj, model)
        assert obj.id == index


def test_bulk_upsert_accepts_getitem_iterable():
    """Tests whether an iterable only implementing the __getitem__ method works
    correctly."""

    class GetItemIterable:
        def __init__(self, items):
            self.items = items

        def __getitem__(self, key):
            return self.items[key]

    model = get_fake_model(
        {
            "id": models.BigAutoField(primary_key=True),
            "name": models.CharField(max_length=255, unique=True),
        }
    )

    rows = GetItemIterable([dict(name="John Smith"), dict(name="Jane Doe")])

    objs = model.objects.bulk_upsert(
        conflict_target=["name"], rows=rows, return_model=True
    )

    for index, obj in enumerate(objs, 1):
        assert isinstance(obj, model)
        assert obj.id == index


def test_bulk_upsert_accepts_iter_iterable():
    """Tests whether an iterable only implementing the __iter__ method works
    correctly."""

    class IterIterable:
        def __init__(self, items):
            self.items = items

        def __iter__(self):
            return iter(self.items)

    model = get_fake_model(
        {
            "id": models.BigAutoField(primary_key=True),
            "name": models.CharField(max_length=255, unique=True),
        }
    )

    rows = IterIterable([dict(name="John Smith"), dict(name="Jane Doe")])

    objs = model.objects.bulk_upsert(
        conflict_target=["name"], rows=rows, return_model=True
    )

    for index, obj in enumerate(objs, 1):
        assert isinstance(obj, model)
        assert obj.id == index



from typing import List, Tuple, Union

from django.contrib.postgres.fields import HStoreField as DjangoHStoreField
from django.db.models.expressions import Expression
from django.db.models.fields import Field


class HStoreField(DjangoHStoreField):
    """Improved version of Django's :see:HStoreField that adds support for
    database-level constraints.

    Notes:
        - For the implementation of uniqueness, see the
          custom database back-end.
    """

    def __init__(
        self,
        *args,
        uniqueness: List[Union[str, Tuple[str, ...]]] = None,
        required: List[str] = None,
        **kwargs
    ):
        """Initializes a new instance of :see:HStoreField.

        Arguments:
            uniqueness:
                List of keys to enforce as unique. Use tuples
                to enforce multiple keys together to be unique.

            required:
                List of keys that should be enforced as required.
        """

        super(HStoreField, self).__init__(*args, **kwargs)

        self.uniqueness = uniqueness
        self.required = required

    def get_prep_value(self, value):
        """Override the base class so it doesn't cast all values to strings.

        psqlextra supports expressions in hstore fields, so casting all
        values to strings is a bad idea.
        """

        value = Field.get_prep_value(self, value)

        if isinstance(value, dict):
            prep_value = {}
            for key, val in value.items():
                if isinstance(val, Expression):
                    prep_value[key] = val
                elif val is not None:
                    prep_value[key] = str(val)
                else:
                    prep_value[key] = val

            value = prep_value

        if isinstance(value, list):
            value = [str(item) for item in value]

        return value

    def deconstruct(self):
        """Gets the values to pass to :see:__init__ when re-creating this
        object."""

        name, path, args, kwargs = super(HStoreField, self).deconstruct()

        if self.uniqueness is not None:
            kwargs["uniqueness"] = self.uniqueness

        if self.required is not None:
            kwargs["required"] = self.required

        return name, path, args, kwargs



from psqlextra.fields import HStoreField


class HStoreRequiredSchemaEditorSideEffect:
    sql_hstore_required_create = (
        "ALTER TABLE {table} "
        "ADD CONSTRAINT {name} "
        "CHECK (({field}->'{key}') "
        "IS NOT NULL)"
    )

    sql_hstore_required_rename = (
        "ALTER TABLE {table} "
        "RENAME CONSTRAINT "
        "{old_name} "
        "TO "
        "{new_name}"
    )

    sql_hstore_required_drop = (
        "ALTER TABLE {table} " "DROP CONSTRAINT IF EXISTS {name}"
    )

    def create_model(self, model):
        """Ran when a new model is created."""

        for field in model._meta.local_fields:
            if not isinstance(field, HStoreField):
                continue

            self.add_field(model, field)

    def delete_model(self, model):
        """Ran when a model is being deleted."""

        for field in model._meta.local_fields:
            if not isinstance(field, HStoreField):
                continue

            self.remove_field(model, field)

    def alter_db_table(self, model, old_db_table, new_db_table):
        """Ran when the name of a model is changed."""

        for field in model._meta.local_fields:
            if not isinstance(field, HStoreField):
                continue

            for key in self._iterate_required_keys(field):
                self._rename_hstore_required(
                    old_db_table, new_db_table, field, field, key
                )

    def add_field(self, model, field):
        """Ran when a field is added to a model."""

        for key in self._iterate_required_keys(field):
            self._create_hstore_required(model._meta.db_table, field, key)

    def remove_field(self, model, field):
        """Ran when a field is removed from a model."""

        for key in self._iterate_required_keys(field):
            self._drop_hstore_required(model._meta.db_table, field, key)

    def alter_field(self, model, old_field, new_field, strict=False):
        """Ran when the configuration on a field changed."""

        is_old_field_hstore = isinstance(old_field, HStoreField)
        is_new_field_hstore = isinstance(new_field, HStoreField)

        if not is_old_field_hstore and not is_new_field_hstore:
            return

        old_required = getattr(old_field, "required", []) or []
        new_required = getattr(new_field, "required", []) or []

        # handle field renames before moving on
        if str(old_field.column) != str(new_field.column):
            for key in self._iterate_required_keys(old_field):
                self._rename_hstore_required(
                    model._meta.db_table,
                    model._meta.db_table,
                    old_field,
                    new_field,
                    key,
                )

        # drop the constraints for keys that have been removed
        for key in old_required:
            if key not in new_required:
                self._drop_hstore_required(model._meta.db_table, old_field, key)

        # create new constraints for keys that have been added
        for key in new_required:
            if key not in old_required:
                self._create_hstore_required(
                    model._meta.db_table, new_field, key
                )

    def _create_hstore_required(self, table_name, field, key):
        """Creates a REQUIRED CONSTRAINT for the specified hstore key."""

        name = self._required_constraint_name(table_name, field, key)

        sql = self.sql_hstore_required_create.format(
            name=self.quote_name(name),
            table=self.quote_name(table_name),
            field=self.quote_name(field.column),
            key=key,
        )
        self.execute(sql)

    def _rename_hstore_required(
        self, old_table_name, new_table_name, old_field, new_field, key
    ):
        """Renames an existing REQUIRED CONSTRAINT for the specified hstore
        key."""

        old_name = self._required_constraint_name(
            old_table_name, old_field, key
        )
        new_name = self._required_constraint_name(
            new_table_name, new_field, key
        )

        sql = self.sql_hstore_required_rename.format(
            table=self.quote_name(new_table_name),
            old_name=self.quote_name(old_name),
            new_name=self.quote_name(new_name),
        )
        self.execute(sql)

    def _drop_hstore_required(self, table_name, field, key):
        """Drops a REQUIRED CONSTRAINT for the specified hstore key."""

        name = self._required_constraint_name(table_name, field, key)

        sql = self.sql_hstore_required_drop.format(
            table=self.quote_name(table_name), name=self.quote_name(name)
        )
        self.execute(sql)

    @staticmethod
    def _required_constraint_name(table: str, field, key):
        """Gets the name for a CONSTRAINT that applies to a single hstore key.

        Arguments:
            table:
                The name of the table the field is
                a part of.

            field:
                The hstore field to create a
                UNIQUE INDEX for.

            key:
                The name of the hstore key
                to create the name for.

        Returns:
            The name for the UNIQUE index.
        """

        return "{table}_{field}_required_{postfix}".format(
            table=table, field=field.column, postfix=key
        )

    @staticmethod
    def _iterate_required_keys(field):
        """Iterates over the keys marked as "required" in the specified field.

        Arguments:
            field:
                The field of which key's to
                iterate over.
        """

        required_keys = getattr(field, "required", None)
        if not required_keys:
            return

        for key in required_keys:
            yield key



# comment placed on partition tables created by the partitioner
# partition tables that do not have this comment will _never_
# be deleted by the partitioner, this is a safety mechanism so
# manually created partitions aren't accidently cleaned up
AUTO_PARTITIONED_COMMENT = "psqlextra_auto_partitioned"


__all__ = ["AUTO_PARTITIONED_COMMENT"]
